from typing import List
from typing import Union

import numpy as np
import pandas as pd

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs


def class_summary(mixdb: MixtureDatabase,
                  mixids: MixtureIDs,
                  truth_f: np.ndarray,
                  predict: np.ndarray,
                  pred_thr: Union[float, List[float]] = 0,
                  truth_thr: float = 0.5,
                  timesteps: int = 0) -> pd.DataFrame:
    """ Calculate table of metrics per class, and averages for a list
        of mixtures using truth and prediction data features x num_classes
        Example:
        Generate multi-class metric summary into table, for example:
                        PPV     TPR      F1     FPR     ACC   AP  AUC  Support
            Class 1     0.71    0.80    0.75    0.00    0.99            44
            Class 2     0.90    0.76    0.82    0.00    0.99            128
            Class 3     0.86    0.82    0.84    0.04    0.93            789
            Other       0.94    0.96    0.95    0.18    0.92            2807

          micro-avg                     0.92    0.027                   3768
          macro avg     0.85    0.83    0.84    0.05    0.96            3768
          micro-avgwo
    """
    from sonusai.metrics import one_hot
    from sonusai.mixture import get_mixids_data

    num_classes = truth_f.shape[1]

    y_truth_f, y_predict = get_mixids_data(mixdb, mixids, truth_f, predict)

    if not mixdb.truth_mutex and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            pred_thr = 0.5

        if np.ndim(pred_thr) == 1 and len(pred_thr) == 1:
            if pred_thr[0] == 0:
                pred_thr = 0.5

    _, metrics, _, _, _, metavg = one_hot(y_truth_f, y_predict, pred_thr, truth_thr, timesteps)

    # [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
    table_idx = np.array([2, 1, 6, 4, 0, 12, 13, 9])
    col_n = ['PPV', 'TPR', 'F1', 'FPR', 'ACC', 'AP', 'AUC', 'Support']
    if mixdb.truth_mutex:
        if len(mixdb.class_labels) >= num_classes - 1:  # labels exist with or without Other
            row_n = mixdb.class_labels
            if len(mixdb.class_labels) == num_classes - 1:  # Other label does not exist, so add it
                row_n.append('Other')
        else:
            row_n = ([f'Class {i}' for i in range(1, num_classes)])
            row_n.append('Other')
    else:
        if len(mixdb.class_labels) == num_classes:
            row_n = mixdb.class_labels
        else:
            row_n = ([f'Class {i}' for i in range(1, num_classes + 1)])

    df = pd.DataFrame(metrics[:, table_idx], columns=col_n, index=row_n)

    # [miPPV, miTPR, miF1, miFPR, miACC, miAP, miAUC, TPSUM]
    avg_row_n = ['Macro-avg', 'Micro-avg', 'Weighted-avg']
    dfavg = pd.DataFrame(metavg, columns=col_n, index=avg_row_n)

    # dfblank = pd.DataFrame([''])
    # pd.concat([df, dfblank, dfblank, dfavg])

    classdf = pd.concat([df, dfavg])
    # classdf = classdf.round(2)
    classdf['Support'] = classdf['Support'].astype(int)

    return classdf



from typing import List

import numpy as np

from sonusai.mixture.mixdb import ClassCount
from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MRecords
from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs


def get_class_count_from_mrecord(mixdb: MixtureDatabase,
                                 mrecord: MRecord,
                                 target_audios: List[np.ndarray],
                                 noise_audio: np.ndarray) -> (ClassCount, np.ndarray):
    """Computes the number of samples for which each truth index is active for a given sample-based truth input."""
    from sonusai.mixture import get_class_weights_threshold
    from sonusai.mixture import generate_truth

    truth_t = generate_truth(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audios=target_audios,
                             noise_audio=noise_audio)

    class_weights_threshold = get_class_weights_threshold(mixdb)

    class_count = [0] * mixdb.num_classes
    num_classes = mixdb.num_classes
    if mixdb.truth_mutex:
        num_classes -= 1
    for cl in range(num_classes):
        class_count[cl] = int(np.sum(truth_t[:, cl] >= class_weights_threshold[cl]))

    return class_count, truth_t


def get_class_count_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> ClassCount:
    """Sums the class counts for given mixids."""
    from sonusai.mixture import get_mrecords_from_mixids

    mixtures = get_mrecords_from_mixids(mixdb, mixids)
    return get_class_count_from_mixtures(mixdb, mixtures)


def get_class_count_from_mixtures(mixdb: MixtureDatabase, mrecords: MRecords) -> ClassCount:
    """Sums the class counts for given mixtures."""
    from sonusai import SonusAIError

    total_class_count = [0] * mixdb.num_classes
    for mrecord in mrecords:
        for cl in range(mixdb.num_classes):
            total_class_count[cl] += mrecord.class_count[cl]

    if mixdb.truth_mutex:
        # Compute the class count for the 'other' class
        if total_class_count[-1] != 0:
            raise SonusAIError('Error: truth_mutex was set, but the class count for the last count was non-zero.')
        total_class_count[-1] = sum([sub.samples for sub in mrecords]) - sum(total_class_count)

    return total_class_count



from typing import Union

import numpy as np

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs


def calculate_class_weights_from_truth(truth: np.ndarray,
                                       other_weight: float = None,
                                       other_index: int = -1) -> np.ndarray:
    """Calculate class weights.

    Supports non-existent classes (a problem with sklearn) where non-existent
    classes get a weight of 0 (instead of inf).
    Includes optional weighting of an "other" class if specified.

    Reference:
        weights = class_weight.compute_class_weight(class_weight='balanced', classes=clabels, y=labels)

    Arguments:
        truth: Truth data in one-hot format. Size can be:
          - frames x timesteps x num_classes
          - frames x num_classes
        other_weight: float or `None`. Weight of the "other" class.
            > 1 = increase weighting/importance relative to the true count
            0 > `other_weight` < 1 = decrease weighting/importance relative
            < 0 or `None` = disable, use true count (default = `None`)
        other_index: int. Index of the "other" class in one-hot mode. Defaults to -1 (the last).

    Returns:
        A numpy array containing class weights.
    """
    frames, num_classes = truth.shape

    if num_classes > 1:
        labels = np.argmax(truth, axis=-1)  # frames x 1 labels from one-hot, last dim
        count = np.bincount(labels, minlength=num_classes).astype(float)
    else:
        num_classes = 2
        labels = np.int8(truth >= 0.5)[:, 0]  # quantize to binary and shape (frames,) for bincount
        count = np.bincount(labels, minlength=num_classes).astype(float)

    if other_weight is not None and other_weight > 0:
        count[other_index] = count[other_index] / np.single(other_weight)

    weights = np.empty((len(count)), dtype=np.single)
    for n in range(len(weights)):
        if count[n] == 0:
            # Avoid sklearn problem with absent classes and assign non-existent classes a weight of 0.
            weights[n] = 0
        else:
            weights[n] = frames / (num_classes * count[n])

    return weights


def calculate_class_weights_from_mixdb(mixdb: MixtureDatabase,
                                       mixids: MixtureIDs = None,
                                       other_weight: Union[float, None] = 1,
                                       other_index: int = -1) -> (np.ndarray, np.ndarray):
    """Calculate class weights using estimated feature counts from a mixture database.

    Arguments:
        mixdb: Mixture database.
        mixids: Mixture ID's.
        other_weight: float or `None`. Weight of the "other" class.
            > 1 = increase weighting/importance relative to the true count
            0 > `other_weight` < 1 = decrease weighting/importance relative
            < 0 or `None` = disable, use true count
        other_index: int. Index of the "other" class in one-hot mode. Defaults to -1 (the last).

    Returns:
        count: Count of features in each class.
        weights: Class weights. num_classes x 1
            Note: for Keras use dict(enumerate(weights))
    """
    from sonusai.mixture import get_class_count_from_mixids

    count = np.ceil(np.array(get_class_count_from_mixids(mixdb=mixdb, mixids=mixids)) / mixdb.feature_step_samples)
    if mixdb.truth_mutex and other_weight is not None and other_weight > 0:
        count[other_index] = count[other_index] / np.single(other_weight)
    total_features = sum(count)

    weights = np.empty(mixdb.num_classes, dtype=np.single)
    for n in range(len(weights)):
        if count[n] == 0:
            # Avoid sklearn problem with absent classes and assign non-existent classes a weight of 0.
            weights[n] = 0
        else:
            weights[n] = total_features / (mixdb.num_classes * count[n])

    return count, weights



from typing import Union

from sonusai.mixture.mixdb import Augmentation
from sonusai.mixture.mixdb import MixtureDatabase


def get_class_balancing_augmentation(mixdb: MixtureDatabase,
                                     target_file_index: int) -> Union[Augmentation, None]:
    """Get the class balancing augmentation rule for the given target."""
    class_balancing_augmentation = mixdb.class_balancing_augmentation
    if mixdb.targets[target_file_index].class_balancing_augmentation is not None:
        class_balancing_augmentation = mixdb.targets[target_file_index].class_balancing_augmentation
    return class_balancing_augmentation



from dataclasses import dataclass
from os import PathLike
from typing import List
from typing import Union

import numpy as np

from sonusai.mixture import DEFAULT_FRAME_SIZE
from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs

MP_DICT = {}


def _get_ft_from_dir_kernel(mixid: int) -> (np.ndarray, np.ndarray):
    """Get feature/truth for a given mixid from a directory containing genft data"""
    from os.path import join

    import h5py

    from sonusai.mixture import get_mixid_padded_name

    mixid_padded_name = get_mixid_padded_name(MP_DICT['mixdb'], mixid)

    with h5py.File(join(MP_DICT['location'], mixid_padded_name + '.h5'), 'r') as f:
        feature = np.array(f['feature'])
        truth_f = np.array(f['truth_f'])

    return feature, truth_f


def get_ft_from_dir(mixdb: MixtureDatabase,
                    location: Union[str, bytes, PathLike],
                    mixids: MixtureIDs = None) -> (np.ndarray, np.ndarray):
    """Get feature/truth for given mixids from a directory containing genft data"""
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.utils import p_map

    MP_DICT['mixdb'] = mixdb
    MP_DICT['location'] = location
    mixids = convert_mixids_to_list(mixdb, mixids)
    result = p_map(_get_ft_from_dir_kernel, mixids)

    feature = np.vstack([result[i][0] for i in range(len(result))])
    truth_f = np.vstack([result[i][1] for i in range(len(result))])

    return feature, truth_f


def get_feature_from_audio(audio: np.ndarray, feature: str) -> np.ndarray:
    from sonusai.mixture import get_pad_length
    from sonusai.mixture import MRecord

    fs = get_feature_stats(feature=feature, num_classes=1)
    audio = np.pad(array=audio,
                   pad_width=(0, get_pad_length(len(audio), fs.feature_step_samples)),
                   mode='constant',
                   constant_values=0)
    mixdb = MixtureDatabase(feature=feature,
                            mixtures=[MRecord(samples=len(audio))],
                            feature_samples=fs.feature_samples,
                            feature_step_samples=fs.feature_step_samples,
                            frame_size=DEFAULT_FRAME_SIZE,
                            num_classes=1,
                            truth_mutex=False,
                            truth_reduction_function='max')
    feature, _, _ = get_feature_and_truth_f(mixdb=mixdb,
                                            mrecord=mixdb.mixtures[0],
                                            mixture=audio,
                                            compute_truth=False)
    return feature


def read_feature_data(filename: str) -> (MixtureDatabase, np.ndarray, np.ndarray, np.ndarray):
    """Read mixdb, feature, truth_f, and segsnr data from given HDF5 file and return them as a tuple."""
    import h5py

    from sonusai.mixture import convert_json_to_mixdb

    with h5py.File(filename, 'r') as f:
        return (convert_json_to_mixdb(f.attrs['mixdb']),
                np.array(f['/feature']),
                np.array(f['/truth_f']),
                np.array(f['/segsnr']))


def get_feature_and_truth_f(mixdb: MixtureDatabase,
                            mrecord: MRecord,
                            mixture: np.ndarray = None,
                            truth_t: np.ndarray = None,
                            segsnr_t: np.ndarray = None,
                            raw_target_audios: List[np.ndarray] = None,
                            augmented_noise_audios: List[List[np.ndarray]] = None,
                            compute_truth: bool = True,
                            compute_segsnr: bool = False) -> (np.ndarray, np.ndarray, np.ndarray):
    from pyaaware import FeatureGenerator
    from pyaaware import ForwardTransform

    from sonusai import SonusAIError
    from sonusai.mixture import get_audio_and_truth_t
    from sonusai.mixture import get_feature_frames_in_mrecord
    from sonusai.mixture import get_transform_frames_in_mrecord
    from sonusai.mixture import truth_reduction
    from sonusai.utils import int16_to_float

    segsnr = None
    if mixture is None or (compute_truth and truth_t is None) or (compute_segsnr and segsnr_t is None):
        (mixture,
         truth_t,
         _,
         _,
         segsnr) = get_audio_and_truth_t(mixdb=mixdb,
                                         mrecord=mrecord,
                                         raw_target_audios=raw_target_audios,
                                         augmented_noise_audios=augmented_noise_audios,
                                         compute_truth=compute_truth,
                                         compute_segsnr=compute_segsnr,
                                         frame_based_segsnr=True)

    if len(mixture) != mrecord.samples:
        raise SonusAIError(f'Wrong number of samples in mixture')

    fft = ForwardTransform(N=mixdb.frame_size * 4,
                           R=mixdb.frame_size)

    fg = FeatureGenerator(frame_size=mixdb.frame_size,
                          feature_mode=mixdb.feature,
                          num_classes=mixdb.num_classes,
                          truth_mutex=mixdb.truth_mutex)

    transform_frames = get_transform_frames_in_mrecord(mixdb, mrecord)
    feature_frames = get_feature_frames_in_mrecord(mixdb, mrecord)

    if truth_t is None:
        truth_t = np.zeros((mrecord.samples, mixdb.num_classes), dtype=np.single)

    feature = np.empty((feature_frames, fg.stride, fg.num_bands), dtype=np.single)
    truth_f = np.empty((feature_frames, mixdb.num_classes), dtype=np.single)

    feature_frame = 0
    for transform_frame in range(transform_frames):
        indices = slice(transform_frame * mixdb.frame_size,
                        (transform_frame + 1) * mixdb.frame_size)
        fd = fft.execute(int16_to_float(mixture[indices]))

        fg.execute(fd, truth_reduction(truth_t[indices], mixdb.truth_reduction_function))

        if fg.eof():
            feature[feature_frame] = fg.feature()
            truth_f[feature_frame] = fg.truth()
            feature_frame += 1

    return feature, truth_f, segsnr


@dataclass(frozen=True)
class FeatureStats:
    feature_ms: float
    feature_samples: int
    feature_step_ms: float
    feature_step_samples: int
    num_bands: int
    stride: int
    step: int
    decimation: int


def get_feature_stats(feature: str,
                      num_classes: int,
                      frame_size: int = DEFAULT_FRAME_SIZE) -> FeatureStats:
    import h5py
    from pyaaware import FeatureGenerator

    import sonusai

    fg = FeatureGenerator(feature_mode=feature,
                          frame_size=frame_size,
                          num_classes=num_classes)

    transform_frame_ms = float(frame_size) / float(sonusai.mixture.SAMPLE_RATE / 1000)

    return FeatureStats(feature_ms=transform_frame_ms * fg.decimation * fg.stride,
                        feature_samples=frame_size * fg.decimation * fg.stride,
                        feature_step_ms=transform_frame_ms * fg.decimation * fg.step,
                        feature_step_samples=frame_size * fg.decimation * fg.step,
                        num_bands=fg.num_bands,
                        stride=fg.stride,
                        step=fg.step,
                        decimation=fg.decimation)



"""sonusai genft

usage: genft [-hvs] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -o OUTPUT, --output OUTPUT      Output directory.
    -s, --segsnr                    Save segsnr. [default: False].

Generate SonusAI feature/truth data from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.
    MIXID       A glob of mixture ID(s) to generate.

Outputs:
    OUTPUT/     A directory containing:
                    <id>.h5:
                        dataset:    feature
                        dataset:    truth_f
                        dataset:    segsnr (optional)
                    <id>.txt
                    genft.log

"""
from typing import List

import numpy as np

from sonusai import logger
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import MixtureIDs

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = {}


def genft(mixdb: MixtureDatabase,
          mixids: MixtureIDs,
          raw_target_audios: List[np.ndarray] = None,
          augmented_noise_audios: List[List[np.ndarray]] = None,
          mixture: np.ndarray = None,
          truth_t: np.ndarray = None,
          segsnr_t: np.ndarray = None,
          compute_truth: bool = True,
          compute_segsnr: bool = False) -> (np.ndarray,
                                            np.ndarray,
                                            np.ndarray):
    import multiprocessing as mp

    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import extract_results
    from sonusai.mixture import get_feature_and_truth_f
    from sonusai.utils import p_map

    mixids = convert_mixids_to_list(mixdb, mixids)
    results = []
    if mp.current_process().daemon:
        for mixid in mixids:
            results.append(get_feature_and_truth_f(mixdb=mixdb,
                                                   mrecord=mixdb.mixtures[mixid],
                                                   mixture=mixture,
                                                   truth_t=truth_t,
                                                   segsnr_t=segsnr_t,
                                                   raw_target_audios=raw_target_audios,
                                                   augmented_noise_audios=augmented_noise_audios,
                                                   compute_truth=compute_truth,
                                                   compute_segsnr=compute_segsnr))
    else:
        MP_DICT['mixdb'] = mixdb
        MP_DICT['mixture'] = mixture
        MP_DICT['truth_t'] = truth_t
        MP_DICT['segsnr_t'] = segsnr_t
        MP_DICT['raw_target_audios'] = raw_target_audios
        MP_DICT['augmented_noise_audios'] = augmented_noise_audios
        MP_DICT['compute_truth'] = compute_truth
        MP_DICT['compute_segsnr'] = compute_segsnr

        results = p_map(_genft_kernel, mixids)

    return extract_results(results)


def _genft_kernel(mixid: int) -> None:
    from sonusai.mixture import get_feature_and_truth_f

    return get_feature_and_truth_f(mixdb=MP_DICT['mixdb'],
                                   mrecord=MP_DICT['mixdb'].mixtures[mixid],
                                   raw_target_audios=MP_DICT['raw_target_audios'],
                                   augmented_noise_audios=MP_DICT['augmented_noise_audios'],
                                   mixture=MP_DICT['mixture'],
                                   truth_t=MP_DICT['truth_t'],
                                   segsnr_t=MP_DICT['segsnr_t'],
                                   compute_truth=MP_DICT['compute_truth'],
                                   compute_segsnr=MP_DICT['compute_segsnr'])


def _process_mixture(mixid: int) -> None:
    from os.path import join

    import h5py

    from sonusai.mixture import get_mixid_padded_name
    from sonusai.mixture import get_mixture_metadata

    mixdb = MP_DICT['mixdb']
    compute_segsnr = MP_DICT['compute_segsnr']
    mixid_padded_name = get_mixid_padded_name(mixdb, mixid)
    output_base = join(MP_DICT['output_dir'], mixid_padded_name)

    with h5py.File(output_base + '.h5', 'a') as f:
        mixture_t = None
        if 'mixture' in f:
            mixture_t = np.array(f['mixture'])

        truth_t = None
        if 'truth_t' in f:
            truth_t = np.array(f['truth_t'])

        segsnr_t = None
        if compute_segsnr and 'segsnr_t' in f:
            segsnr_t = np.array(f['segsnr_t'])

        feature, truth_f, segsnr = genft(mixdb=mixdb,
                                         mixids=mixid,
                                         raw_target_audios=MP_DICT['raw_target_audios'],
                                         augmented_noise_audios=MP_DICT['augmented_noise_audios'],
                                         mixture=mixture_t,
                                         truth_t=truth_t,
                                         segsnr_t=segsnr_t,
                                         compute_segsnr=compute_segsnr)

        datasets = ['feature', 'truth_f', 'segsnr']
        for dataset in datasets:
            if dataset in f:
                del f[dataset]

        f.create_dataset(name='feature', data=feature, dtype=np.single)
        f.create_dataset(name='truth_f', data=truth_f, dtype=np.single)
        if compute_segsnr:
            f.create_dataset(name='segsnr', data=segsnr, dtype=np.single)

    with open(file=output_base + '.txt', mode='w') as f:
        f.write(get_mixture_metadata(mixdb, mixdb.mixtures[mixid]))


def main():
    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from docopt import docopt
    from pyaaware import FeatureGenerator
    from tqdm import tqdm

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import get_augmented_noise_audios
    from sonusai.mixture import get_raw_audios
    from sonusai.mixture import check_audio_files_exist
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import load_mixdb
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixdb_name = args['--mixdb']
    mixid = args['--mixid']
    output_dir = args['--output']
    compute_segsnr = args['--segsnr']

    if not output_dir:
        output_dir = splitext(mixdb_name)[0]

    if exists(output_dir) and not isdir(output_dir):
        remove(output_dir)

    makedirs(output_dir, exist_ok=True)

    start_time = time.monotonic()

    create_file_handler(join(output_dir, 'genft.log'))
    update_console_handler(verbose)
    initial_log_messages('genft')

    logger.info(f'\nLoad mixture database from {mixdb_name}')
    mixdb = load_mixdb(name=mixdb_name)
    mixid = convert_mixids_to_list(mixdb, mixid)

    total_samples = sum([sub.samples for sub in [mixdb.mixtures[m] for m in mixid]])
    duration = total_samples / sonusai.mixture.SAMPLE_RATE
    total_transform_frames = total_samples // mixdb.frame_size
    total_feature_frames = total_samples // mixdb.feature_step_samples

    logger.info('')
    logger.info(f'Found {len(mixid):,} mixtures to process')
    logger.info(f'{total_samples:,} samples, '
                f'{total_transform_frames:,} transform frames, '
                f'{total_feature_frames:,} feature frames')

    check_audio_files_exist(mixdb)

    fg = FeatureGenerator(frame_size=mixdb.frame_size,
                          feature_mode=mixdb.feature,
                          num_classes=mixdb.num_classes,
                          truth_mutex=mixdb.truth_mutex)

    MP_DICT['mixdb'] = mixdb
    MP_DICT['output_dir'] = output_dir
    MP_DICT['raw_target_audios'] = get_raw_audios(mixdb)
    MP_DICT['augmented_noise_audios'] = get_augmented_noise_audios(mixdb)
    MP_DICT['compute_segsnr'] = compute_segsnr

    progress = tqdm(total=len(mixid), desc='genft')
    p_tqdm_map(_process_mixture, mixid, progress=progress)
    progress.close()

    logger.info(f'Wrote {len(mixid)} mixtures to {output_dir}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'feature:  {human_readable_size(total_feature_frames * fg.stride * fg.num_bands * 4, 1)}')
    logger.info(f'truth_f:  {human_readable_size(total_feature_frames * mixdb.num_classes * 4, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(total_transform_frames * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



def yes_or_no(question: str) -> bool:
    """Wait for yes or no input"""
    while 'the answer is invalid':
        reply = str(input(question + ' (y/n)?: ')).lower().strip()
        if reply[:1] == 'y':
            return True
        if reply[:1] == 'n':
            return False



from typing import Any

from sonusai import logger


def keras_onnx(model_name: str,
               weight_name: str,
               timesteps: int = None,
               batch_size: int = None,
               output_dir: str = None) -> None:
    from os.path import basename
    from os.path import dirname
    from os.path import join
    from os.path import splitext

    import h5py
    import keras_tuner as kt

    if output_dir is None:
        output_dir = dirname(model_name)

    with h5py.File(weight_name, 'r') as f:
        feature = f.attrs['sonusai_feature']
        num_classes = int(f.attrs['sonusai_num_classes'])

    model = import_keras_model(model_name)
    model_base = basename(model_name)
    model_root = splitext(model_base)[0]
    base_name = join(output_dir, model_root)

    # Build default model to check overrides
    timesteps = check_keras_overrides(model, feature, num_classes, timesteps, batch_size)

    try:
        hypermodel = model.MyHyperModel(feature=feature,
                                        num_classes=num_classes,
                                        timesteps=timesteps,
                                        batch_size=batch_size)
        built_model = hypermodel.build_model(kt.HyperParameters())
    except Exception as e:
        logger.exception(f'Error: build_model() in {model_base} failed: {e}.')
        raise SystemExit(1)

    print('')
    print(f'timesteps               {timesteps}')
    print(f'batch_size              {batch_size}')
    print(f'hypermodel.timesteps    {hypermodel.timesteps}')
    print(f'hypermodel.batch_size   {hypermodel.batch_size}')
    print('')

    # Create and save ONNX model with specified batch and timestep sizes
    onnx_name = base_name + '.onnx'
    try:
        create_onnx_from_keras(keras_model=built_model,
                               is_flattened=hypermodel.flatten,
                               has_timestep=(hypermodel.timesteps > 0),
                               has_channel=hypermodel.add1ch,
                               is_mutex=hypermodel.truth_mutex,
                               feature=feature,
                               filename=onnx_name)
    except Exception as e:
        logger.warning(f'Failed to create ONNX model, no file saved: {e}.')
    logger.info(f'Wrote ONNX model to {onnx_name}')

    # Create and save onnx model with timesteps, batch = 1
    timesteps = hypermodel.timesteps
    if timesteps > 0:
        # only set to 1 if nonzero (exists)
        timesteps = 1
    hypermodel = model.MyHyperModel(feature=feature,
                                    num_classes=num_classes,
                                    timesteps=timesteps,
                                    batch_size=1)
    built_model = hypermodel.build_model(kt.HyperParameters())
    # load weights from previously saved HDF5
    built_model.load_weights(weight_name)
    # save a prediction version of model to base_name-pred-onnx
    onnx_name = base_name + '-b1.onnx'
    create_onnx_from_keras(keras_model=built_model,
                           is_flattened=hypermodel.flatten,
                           has_timestep=(hypermodel.timesteps > 0),
                           has_channel=hypermodel.add1ch,
                           is_mutex=hypermodel.truth_mutex,
                           feature=feature,
                           filename=onnx_name)
    logger.info(f'Wrote inference ONNX model to {onnx_name}')


def check_keras_overrides(model: Any, feature: str, num_classes: int, timesteps: int, batch_size: int) -> int:
    try:
        hypermodel = model.MyHyperModel()
    except Exception as e:
        logger.exception(f'Error: build_model() failed: {e}.')
        raise SystemExit(1)

    if hypermodel.feature != feature:
        logger.warning(f'Overriding feature: default = {hypermodel.feature}; specified = {feature}.')

    if hypermodel.num_classes != num_classes:
        logger.warning(
            f'Overriding num_classes: default = {hypermodel.num_classes}; specified = {num_classes}.')

    if timesteps is not None:
        if hypermodel.timesteps == 0 and timesteps != 0:
            logger.warning(f'Model does not contain timesteps; ignoring override.')
            timesteps = 0

        if hypermodel.timesteps != 0 and timesteps == 0:
            logger.warning(f'Model contains timesteps; ignoring override.')
            timesteps = hypermodel.timesteps

        if hypermodel.timesteps != timesteps:
            logger.info(f'Overriding timesteps: default = {hypermodel.timesteps}; specified = {timesteps}.')

    if batch_size is not None and hypermodel.batch_size != batch_size:
        logger.info(f'Overriding batch_size: default = {hypermodel.batch_size}; specified = {batch_size}.')

    return timesteps


def import_keras_model(name: str) -> Any:
    import os
    import sys
    from importlib import import_module

    from sonusai import SonusAIError

    try:
        path = os.path.dirname(name)
        if len(path) < 1:
            path = './'

        # Add model file location to system path
        sys.path.append(os.path.abspath(path))

        try:
            # expect at least a Keras-tuner build_model(hp) function
            root = os.path.splitext(os.path.basename(name))[0]
            model = import_module(root)
        except Exception as e:
            raise SonusAIError(f'Error: could not import model from {name}: {e}.')
    except Exception as e:
        raise SonusAIError(f'Error: could not find {name}: {e}.')

    return model


def create_onnx_from_keras(keras_model,
                           is_flattened: bool = True,
                           has_timestep: bool = True,
                           has_channel: bool = False,
                           is_mutex: bool = True,
                           feature: str = '',
                           filename: str = ''):
    import tf2onnx
    from tensorflow.keras import backend as kb

    from sonusai.utils import add_sonusai_metadata
    from sonusai.utils import replace_stateful_grus

    kb.clear_session()
    onnx_model, _ = tf2onnx.convert.from_keras(keras_model)
    onnx_model = replace_stateful_grus(keras_model=keras_model, onnx_model=onnx_model)
    onnx_model = add_sonusai_metadata(model=onnx_model,
                                      is_flattened=is_flattened,
                                      has_timestep=has_timestep,
                                      has_channel=has_channel,
                                      is_mutex=is_mutex,
                                      feature=feature)
    if filename:
        import onnx
        onnx.save(onnx_model, filename)
    return onnx_model



"""sonusai keras_predict

usage: keras_predict [-hvr] (-m MODEL) (-w KMODEL) (-d INPUT) [-b BATCH] [-t TSTEPS] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -m MODEL, --model MODEL         Python model file.
    -w KMODEL, --weights KMODEL     Keras model weights file.
    -d INPUT, --input INPUT         Input data.
    -b BATCH, --batch BATCH         Batch size.
    -t TSTEPS, --tsteps TSTEPS      Timesteps.
    -o OUTPUT, --output OUTPUT      Output directory.
    -r, --reset                     Reset model between each file.

Run prediction on a trained Keras model defined by a SonusAI Keras Python model file using SonusAI genft data
and/or WAV data.

Inputs:
    MODEL       A SonusAI Python model file with build and/or hypermodel functions.
    KMODEL      A Keras model weights file (or model file with weights).
    INPUT       The input data can be a file, directory, or glob of genft H5 files and/or WAV files.

Outputs:
    OUTPUT/     A directory containing:
                    <id>.h5
                        dataset:    predict
                    keras_predict.log

Results are written into subdirectory kpredict-<TIMESTAMP> unless OUTPUT is specified.

"""
from sonusai import logger


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    weights_name = args['--weights']
    input_name = args['--input']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    output_dir = args['--output']
    reset = args['--reset']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import h5py
    import keras_tuner as kt
    import numpy as np
    import tensorflow as tf
    from tensorflow.keras import backend as kb

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import SequenceFromH5
    from sonusai.data_generator import get_frames_per_batch
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import read_audio
    from sonusai.utils import braced_glob
    from sonusai.utils import check_keras_overrides
    from sonusai.utils import create_ts_name
    from sonusai.utils import import_keras_model
    from sonusai.utils import reshape_inputs
    from sonusai.utils import reshape_outputs

    model_base = basename(model_name)

    if batch_size is not None:
        batch_size = int(batch_size)

    if timesteps is not None:
        timesteps = int(timesteps)

    if output_dir is None:
        output_dir = create_ts_name('kpredict')

    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'keras_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('keras_predict')

    logger.info(f'tensorflow    {tf.__version__}')
    logger.info(f'keras         {tf.keras.__version__}')
    logger.info('')

    # Import model definition file
    logger.info(f'Importing {model_base}')
    model = import_keras_model(model_name)
    feature = None
    num_classes = None
    with h5py.File(weights_name, 'r') as f:
        if 'sonusai_feature' in f.attrs:
            feature = f.attrs['sonusai_feature']
        else:
            logger.warn(f'Could not find SonusAI feature in weights file; using hypermodel default.')
        if 'sonusai_num_classes' in f.attrs:
            num_classes = int(f.attrs['sonusai_num_classes'])
        else:
            logger.warn(f'Could not find SonusAI num_classes in weights file; using hypermodel default.')

    # Check overrides
    timesteps = check_keras_overrides(model, feature, num_classes, timesteps, batch_size)

    logger.info('Building model')
    try:
        hypermodel = model.MyHyperModel(feature=feature,
                                        num_classes=num_classes,
                                        timesteps=timesteps,
                                        batch_size=batch_size)
        built_model = hypermodel.build_model(kt.HyperParameters())
    except Exception as e:
        logger.exception(f'Error: build_model() in {model_base} failed: {e}.')
        raise SystemExit(1)

    frames_per_batch = get_frames_per_batch(hypermodel.batch_size, hypermodel.timesteps)

    kb.clear_session()
    logger.info('')
    built_model.summary(print_fn=logger.info)
    logger.info('')
    logger.info(f'feature       {hypermodel.feature}')
    logger.info(f'num_classes   {hypermodel.num_classes}')
    logger.info(f'batch_size    {hypermodel.batch_size}')
    logger.info(f'timesteps     {hypermodel.timesteps}')
    logger.info(f'flatten       {hypermodel.flatten}')
    logger.info(f'add1ch        {hypermodel.add1ch}')
    logger.info(f'truth_mutex   {hypermodel.truth_mutex}')
    logger.info(f'lossf         {hypermodel.lossf}')
    logger.info(f'input_shape   {hypermodel.input_shape}')
    logger.info(f'optimizer     {built_model.optimizer.get_config()}')
    logger.info('')

    logger.info(f'Loading weights from {weights_name}')
    built_model.load_weights(weights_name)

    if isdir(input_name):
        input_name = join(input_name, '*.{h5,wav}')

    files = [x for x in braced_glob(input_name) if splitext(x)[1] == '.h5']

    # Convert WAV to feature data
    wav_files = [x for x in braced_glob(input_name) if splitext(x)[1] == '.wav']
    for file in wav_files:
        audio = read_audio(file)
        data = get_feature_from_audio(audio=audio, feature=feature)

        output_base = basename(splitext(file)[0])
        output_name = join(output_dir, output_base)
        if output_base + '.h5' in [basename(x) for x in files]:
            # If an H5 file of the same name as the WAV file already exists, then append '_wav'
            # to the output file name.
            output_name += '_wav'
        output_name += '.h5'
        files.append(output_name)
        # Save the newly created feature
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=data, dtype=np.single)

    files = sorted(files)
    logger.info(f'Found {len(files)} files from {input_name}')

    if reset:
        # reset mode cycles through each file one at a time
        for file in files:
            # Read in H5
            with h5py.File(file, 'r') as f:
                data = np.array(f['feature'])

            # Pad with zeros in order to create an entire batch of data
            frames = data.shape[0]
            padding = frames_per_batch - frames % frames_per_batch
            data = np.pad(array=data,
                          pad_width=((0, padding), (0, 0), (0, 0)),
                          mode='constant',
                          constant_values=0)
            data, _, _, _, _, _ = reshape_inputs(feature=data,
                                                 batch_size=hypermodel.batch_size,
                                                 timesteps=hypermodel.timesteps,
                                                 flatten=hypermodel.flatten,
                                                 add1ch=hypermodel.add1ch)

            predict = built_model.predict(data, batch_size=hypermodel.batch_size, verbose=1)
            predict, _, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)
            predict = predict[:frames, :]

            output_name = join(output_dir, basename(file))
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset(name='predict', data=predict, dtype=np.single)
    else:
        # Run all data at once using a data generator
        data = SequenceFromH5(files=files,
                              feature=feature,
                              num_classes=num_classes,
                              batch_size=hypermodel.batch_size,
                              timesteps=hypermodel.timesteps,
                              flatten=hypermodel.flatten,
                              add1ch=hypermodel.add1ch,
                              truth_mutex=hypermodel.truth_mutex)

        predict = built_model.predict(data, batch_size=hypermodel.batch_size, verbose=1)
        predict, _, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)

        # Write data to separate files
        for idx, file in enumerate(files):
            output_name = join(output_dir, basename(file))
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset('predict', data=predict[data.file_indices[idx]], dtype=np.single)

    logger.info(f'Saved results to {output_dir}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from typing import List

from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MRecords
from sonusai.mixture.mixdb import MixtureDatabase

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = {}


def balance_active_truth(mixdb: MixtureDatabase,
                         raw_target_audio: list,
                         logging: bool = True,
                         show_progress: bool = False) -> MRecords:
    """Add target augmentations until the class count values are balanced."""

    import numpy as np

    from sonusai import logger
    from sonusai.mixture.constants import SAMPLE_RATE
    from sonusai.utils.seconds_to_hms import seconds_to_hms

    MP_DICT['mixdb'] = mixdb
    MP_DICT['raw_target_audio'] = raw_target_audio

    augmented_targets = _get_augmented_targets(mixdb)
    class_balancing_samples = _get_class_balancing_samples(mixdb, augmented_targets)
    if logging:
        logger.info('')
        label_digits = max([len(_get_class_label(mixdb, item)) for item in range(len(class_balancing_samples))])
        samples_digits = np.ceil(np.log10(float(max(class_balancing_samples))))
        samples_digits = int(samples_digits + np.ceil(samples_digits / 3))
        for class_index, required_samples in enumerate(class_balancing_samples):
            logger.info(f'Class {_get_class_label(mixdb, class_index):>{label_digits}} '
                        f'needs {required_samples:>{samples_digits},} more active truth samples '
                        f' - {seconds_to_hms(required_samples / SAMPLE_RATE)}')
        logger.info('')

    for class_index, required_samples in enumerate(class_balancing_samples):
        augmented_targets = _balance_class(mixdb=mixdb,
                                           mrecords=augmented_targets,
                                           class_index=class_index,
                                           required_samples=required_samples,
                                           logging=logging,
                                           show_progress=show_progress)

    return augmented_targets


def _balance_class(mixdb: MixtureDatabase,
                   mrecords: MRecords,
                   class_index: int,
                   required_samples: int,
                   logging: bool = True,
                   show_progress: bool = False) -> MRecords:
    """Add target augmentations for a single class until the required samples are satisfied."""
    import multiprocessing as mp

    import numpy as np
    from tqdm import tqdm

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture.targets import get_target_indices_for_truth_index
    from sonusai.utils.parallel import p_map

    if required_samples == 0:
        return mrecords

    class_label = _get_class_label(mixdb, class_index)

    # Get list of targets for this class
    target_indices = get_target_indices_for_truth_index(mixdb.targets, class_index)
    if not target_indices:
        raise SonusAIError(f'Could not find single-class targets for class index {class_index}')

    num_cpus = mp.cpu_count()

    remaining_samples = required_samples
    added_samples = 0
    added_targets = 0
    progress = tqdm(total=required_samples, desc=f'Balance class {class_label}', disable=not show_progress)
    while True:
        new_mrecords: MRecords = []
        while len(new_mrecords) < num_cpus:
            for target_index in target_indices:
                augmentation_indices = _get_unused_balancing_augmentations(mixdb=mixdb,
                                                                           mrecords=mrecords,
                                                                           target_file_index=target_index,
                                                                           amount=num_cpus)
                for augmentation_index in augmentation_indices:
                    new_mrecords.append(MRecord(target_file_index=[target_index],
                                                target_augmentation_index=[augmentation_index]))

        new_mrecords = new_mrecords[0:num_cpus]
        new_mrecords = p_map(_process_target, new_mrecords)

        for mrecord in new_mrecords:
            new_samples = np.sum(np.sum(mrecord.class_count))
            remaining_samples -= new_samples

            # If the current mrecord will overshoot the required samples then add it only if
            # overshooting results in a sample count closer to the required than not overshooting.
            if remaining_samples >= 0 or -remaining_samples < remaining_samples + new_samples:
                mrecords.append(mrecord)
                added_samples += new_samples
                added_targets += 1
                progress.update(new_samples)

            if remaining_samples <= 0:
                _remove_unused_augmentations(mixdb=mixdb, mrecords=mrecords)
                progress.update(required_samples - added_samples)
                progress.close()
                if logging:
                    logger.info(f'Added {added_targets:,} new augmented targets for class {class_label}')
                return mrecords


def _process_target(mrecord: MRecord) -> MRecord:
    from sonusai.mixture.process import process_target

    return process_target(mrecord=mrecord, mixdb=MP_DICT['mixdb'],
                          raw_target_audios=MP_DICT['raw_target_audio'])


def _get_class_balancing_samples(mixdb: MixtureDatabase, mrecords: MRecords) -> List[int]:
    """Determine the number of additional active truth samples needed for each class in order for
    all classes to be represented evenly over all mixtures.

    If the truth mode is mutually exclusive, ignore the last class (i.e., set to zero).
    """

    import numpy as np

    from sonusai.mixture.class_count import get_class_count_from_mixtures

    class_count = get_class_count_from_mixtures(mixdb, mrecords)

    if mixdb.truth_mutex:
        class_count = class_count[:-1]

    result = list(np.max(class_count) - class_count)

    if mixdb.truth_mutex:
        result.append(0)

    return result


def _get_augmented_targets(mixdb: MixtureDatabase) -> MRecords:
    """Get a list of augmented targets from a mixture database."""
    snr = max(mixdb.snrs)
    return [sub for sub in mixdb.mixtures if sub.snr == snr and sub.noise_file_index == 0]


def _get_class_label(mixdb: MixtureDatabase, class_index: int) -> str:
    if mixdb.class_labels:
        return mixdb.class_labels[class_index]

    return str(class_index)


def _get_unused_balancing_augmentations(mixdb: MixtureDatabase,
                                        mrecords: MRecords,
                                        target_file_index: int,
                                        amount: int = 1) -> List[int]:
    """Get a list of unused balancing augmentations for a given target file index."""
    from dataclasses import asdict

    from sonusai.mixture.augmentation import get_augmentations
    from sonusai.mixture.balance import get_class_balancing_augmentation

    balancing_augmentations = [item for item in range(len(mixdb.target_augmentations)) if
                               item >= mixdb.first_cba_index]
    used_balancing_augmentations = [sub.target_augmentation_index for sub in mrecords if
                                    sub.target_file_index == target_file_index and
                                    sub.target_augmentation_index in balancing_augmentations]

    augmentation_indices = [item for item in balancing_augmentations if item not in used_balancing_augmentations]
    class_balancing_augmentation = get_class_balancing_augmentation(mixdb=mixdb, target_file_index=target_file_index)

    while len(augmentation_indices) < amount:
        new_augmentation = get_augmentations(asdict(class_balancing_augmentation))[0]
        mixdb.target_augmentations.append(new_augmentation)
        augmentation_indices.append(len(mixdb.target_augmentations) - 1)

    return augmentation_indices


def _remove_unused_augmentations(mixdb: MixtureDatabase, mrecords: MRecords) -> None:
    """Remove any unused target augmentation rules from the end of the database."""
    max_used_augmentation = max([index for mixture in mrecords for index in mixture.target_augmentation_index]) + 1
    mixdb.target_augmentations = mixdb.target_augmentations[0:max_used_augmentation]



"""sonusai genmixdb

usage: genmixdb [-hvmfs] CONFIG...

options:
   -h, --help
   -v, --verbose    Be verbose.
   -m, --mix        Save mixture data. [default: False].
   -f, --ft         Save feature/truth_f data. [default: False].
   -s, --segsnr     Save segsnr data. [default: False].

Create mixture database data for training and evaluation. Optionally, also create mixture audio and feature/truth data.

genmixdb creates a database of training and evaluation feature and truth data generation information. It allows
the choice of audio neural-network feature types that are supported by the Aaware real-time front-end and truth
data that is synchronized frame-by-frame with the feature data.

Here are some examples:

#### Adding target data
Suppose you have an audio file which is an example, or target, of what you want to
recognize or detect. Of course, for training a NN you also need truth data for that
file (also called labels). If you don't already have it, genmixdb can create truth using
energy-based sound detection on each frame of the feature data. You can also select
different feature types. Here's an example:

genmixdb target_gfr32ts2.yml

where target_gfr32ts2.yml contains:
---
feature: gfr32ts2

targets:
  - name: data/target.wav

target_augmentations:
  - normalize: -3.5
...

The mixture database is written to a JSON file that inherits the base name of the config file.

#### Target data mix with noise and augmentation

genmixdb mix_gfr32ts2.yml

where mix_gfr32ts2.yml contains:
---
feature: gfr32ts2

output: data/my_mix.h5

targets:
  - name: data/target.wav

target_augmentations:
  - normalize: -3.5
    pitch: [-3, 0, 3]
    tempo: [0.8, 1, 1.2]

noises:
  - name: data/noise.wav

noise_augmentations:
  - normalize: -3.5

snrs:
  - 20
...

In this example a time-domain mixture is created and feature data is calculated as
specified by 'feature: gfr32ts2'. Various feature types are available which vary in
spectral and temporal resolution (4 ms or higher), and other feature algorithm
parameters. The total feature size, dimension, and #frames for mixture is reported
in the log file (the log file name is derived from the output file base name; in this
case it would be mix_gfr32ts2.log).

Truth (labels) can be automatically created per feature output frame based on sound
energy detection. By default, these are appended to the feature data in a single HDF5
output file. By default, truth/label generation is turned on with default algorithm
and threshold levels (see truth section) and a single class, i.e., detecting a single
type of sound. The truth format is a single float per class representing the
probability of activity/presence, and multi-class truth/labels are possible by
specifying the number of classes and either a scalar index or a vector of indices in
which to put the truth result. For example, 'num_class: 3' and  'truth_index: 2' adds
a 1x3 vector to the feature data with truth put in index 2 (others would be 0) for
data/target.wav being an audio clip from sound type of class 2.

The mixture is created with potential data augmentation functions in the following way:
1. apply noise augmentation rule
2. apply target augmentation rule
3. adjust noise gain for specific SNR
4. add augmented noise to augmented target

The mixture length is the target length by default, and the noise signal is repeated
if it is shorter, or trimmed if longer.

#### Target and noise using path lists

Target and noise audio is specified as a list containing text files, audio files, and
file globs. Text files are processed with items on each line where each item can be a
text file, an audio file, or a file glob. Each item will be searched for audio files
which can be WAV, MP3, FLAC, AIFF, or OGG format with any sample rate, bit depth, or
channel count. All audio files will be converted to 16 kHz, 16-bit, single channel
format before processing. For example,

genmixdb dog-bark.yml

where dog-bark.yml contains:
---
targets:
  - name: slib/dog-outside/*.wav
  - name: slib/dog-inside/*.wav

will find all .wav files in the specified directories and process them as targets.

"""
from typing import List

import numpy as np

from sonusai import logger
from sonusai.mixture import AugmentedTargets
from sonusai.mixture import MRecord
from sonusai.mixture import MixtureDatabase

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = {}


def genmixdb(file: str = None,
             config: dict = None,
             save_mix: bool = False,
             save_ft: bool = False,
             save_segsnr: bool = False,
             output_dir: str = None,
             logging: bool = True,
             show_progress: bool = False) -> MixtureDatabase:
    from random import seed

    import yaml
    from tqdm import tqdm

    import sonusai
    from sonusai import SonusAIError
    from sonusai.mixture import Augmentation
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import TruthSettings
    from sonusai.mixture import balance_targets
    from sonusai.mixture import get_augmented_noise_audios
    from sonusai.mixture import evaluate_random_rule
    from sonusai.mixture import get_augmentation_indices_for_mixup
    from sonusai.mixture import get_augmentations
    from sonusai.mixture import get_augmented_target_indices_for_mixup
    from sonusai.mixture import get_augmented_targets
    from sonusai.mixture import get_class_weights_threshold
    from sonusai.mixture import get_feature_stats
    from sonusai.mixture import get_mixups
    from sonusai.mixture import get_noise_files
    from sonusai.mixture import get_raw_audios
    from sonusai.mixture import get_target_files
    from sonusai.mixture import get_class_count_from_mixids
    from sonusai.mixture import load_config
    from sonusai.mixture import log_duration_and_sizes
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import update_truth_settings
    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import seconds_to_hms

    if logging:
        logger.info('')

    if (file is None and config is None) or (file is not None and config is not None):
        raise SonusAIError(f'Must specify either file name or config')

    if file is not None:
        config = load_config(file)

    seed(config['seed'])

    if logging:
        logger.debug(f'Seed: {config["seed"]}')
        logger.debug('Configuration:')
        logger.debug(yaml.dump(config))

    if logging:
        logger.info('Collecting targets')
    target_files = get_target_files(config)
    if len(target_files) == 0:
        raise SonusAIError('Canceled due to no targets')

    if logging:
        logger.debug('List of targets:')
        logger.debug(yaml.dump([sub.name for sub in target_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting noises')
    noise_files = get_noise_files(config)
    if logging:
        logger.debug('List of noises:')
        logger.debug(yaml.dump([sub.name for sub in noise_files], default_flow_style=False))
        logger.debug('')

    if logging:
        logger.info('Collecting target augmentations')
    target_augmentations = get_augmentations(rules=config['target_augmentations'])
    mixups = get_mixups(target_augmentations)
    if logging:
        for mixup in mixups:
            logger.debug(f'Expanded list of target augmentations for mixup of {mixup}:')
            indices = get_augmentation_indices_for_mixup(target_augmentations, mixup)
            for idx in indices:
                logger.debug(f'- {target_augmentations[idx]}')
            logger.debug('')

    if logging:
        logger.info('Collecting noise augmentations')
    noise_augmentations = get_augmentations(config['noise_augmentations'])
    if logging:
        logger.debug('Expanded list of noise augmentations:')
        for augmentation in noise_augmentations:
            logger.debug(f'- {augmentation}')
        logger.debug('')

    if logging:
        logger.debug(f'SNRs: {config["snrs"]}\n')
        logger.debug(f'Random SNRs: {config["random_snrs"]}\n')
        logger.debug(f'Exhaustive noise: {config["exhaustive_noise"]}\n')

    if config['truth_mode'] not in ['normal', 'mutex']:
        raise SonusAIError(f'invalid truth_mode: {config["truth_mode"]}')
    truth_mutex = config['truth_mode'] == 'mutex'

    if truth_mutex and any(mixup > 1 for mixup in mixups):
        raise SonusAIError(f'Mutex truth mode is not compatible with mixup')

    fs = get_feature_stats(feature=config['feature'],
                           num_classes=config['num_classes'],
                           frame_size=config['frame_size'])

    augmented_targets = get_augmented_targets(target_files, target_augmentations)

    mixdb = MixtureDatabase(
        class_balancing=config['class_balancing'],
        class_balancing_augmentation=dataclass_from_dict(Augmentation, config['class_balancing_augmentation']),
        class_labels=config['class_labels'],
        class_weights_threshold=get_class_weights_threshold(config),
        exhaustive_noise=config['exhaustive_noise'],
        feature=config['feature'],
        feature_samples=fs.feature_samples,
        feature_step_samples=fs.feature_step_samples,
        first_cba_index=len(target_augmentations),
        frame_size=config['frame_size'],
        mixtures=[],
        noise_augmentations=noise_augmentations,
        noises=noise_files,
        num_classes=config['num_classes'],
        random_snrs=config['random_snrs'],
        seed=config['seed'],
        snrs=config['snrs'],
        target_augmentations=target_augmentations,
        targets=target_files,
        truth_mutex=truth_mutex,
        truth_reduction_function=config['truth_reduction_function'],
        truth_settings=dataclass_from_dict(TruthSettings, update_truth_settings(config['truth_settings'])),
    )

    if logging:
        logger.info('')

    raw_target_audios = get_raw_audios(mixdb=mixdb, show_progress=show_progress)
    raw_target_audio_samples = sum([len(x) for x in raw_target_audios])
    if logging:
        logger.info('')
        logger.info(f'Raw target audio: {seconds_to_hms(seconds=raw_target_audio_samples / SAMPLE_RATE)}')
        logger.info('')

    augmented_noise_audios = get_augmented_noise_audios(mixdb=mixdb, show_progress=show_progress)
    noise_audio_samples = sum([len(xa) for x in augmented_noise_audios for xa in x])
    if logging:
        logger.info('')
        logger.info(f'Augmented noise audio: {seconds_to_hms(seconds=noise_audio_samples / SAMPLE_RATE)}')
        logger.info('')

    augmented_targets = balance_targets(mixdb, augmented_targets)

    target_sets = 0
    total_duration = 0
    for mixup in mixups:
        augmented_target_indices_for_mixup = get_augmented_target_indices_for_mixup(
            mixdb=mixdb,
            augmented_targets=augmented_targets,
            mixup=mixup)
        target_sets += len(augmented_target_indices_for_mixup)
        for indices in augmented_target_indices_for_mixup:
            for augmented_target in [augmented_targets[idx] for idx in indices]:
                length = int(target_files[augmented_target.target_file_index].duration * sonusai.mixture.SAMPLE_RATE)
                augmentation = target_augmentations[augmented_target.target_augmentation_index]
                if augmentation.tempo is not None:
                    length /= augmentation.tempo
                if length % fs.feature_step_samples:
                    length += fs.feature_step_samples - int(length % fs.feature_step_samples)
                total_duration += float(length) / sonusai.mixture.SAMPLE_RATE

    total_snrs = len(mixdb.snrs) + len(mixdb.random_snrs)
    target_sets *= total_snrs
    total_duration *= total_snrs
    noise_sets = len(noise_files) * len(noise_augmentations) if mixdb.exhaustive_noise else 1
    total_duration *= noise_sets
    total_mixtures = noise_sets * target_sets
    mixdb.mixid_padding = int(np.ceil(np.log10(total_mixtures)))
    if logging:
        logger.info('')
        logger.info(f'Found {total_mixtures:,} mixtures to process')

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb.num_classes,
                               feature_step_samples=fs.feature_step_samples,
                               num_bands=fs.num_bands,
                               stride=fs.stride,
                               desc='Estimated')
        logger.info(f'Feature shape:        {fs.stride} x {fs.num_bands} ({fs.stride * fs.num_bands} total params)')
        logger.info(f'Feature samples:      {fs.feature_samples} samples ({fs.feature_ms} ms)')
        logger.info(f'Feature step samples: {fs.feature_step_samples} samples ({fs.feature_step_ms} ms)')
        logger.info('')

    used_noise_samples = 0
    if mixdb.exhaustive_noise:
        # Get indices and offsets
        for noise_file_index in range(len(noise_files)):
            for noise_augmentation_index, noise_augmentation in enumerate(noise_augmentations):
                noise_offset = 0
                noise_length = len(augmented_noise_audios[noise_file_index][noise_augmentation_index])

                for mixup in mixups:
                    augmented_target_indices_for_mixup = get_augmented_target_indices_for_mixup(
                        mixdb=mixdb,
                        augmented_targets=augmented_targets,
                        mixup=mixup)

                    for augmented_target_index_for_mixup in augmented_target_indices_for_mixup:
                        (target_file_index,
                         target_augmentation_index,
                         target_length) = _get_target_info(
                            mixdb=mixdb,
                            augmented_target_indices_for_mixup=augmented_target_index_for_mixup,
                            augmented_targets=augmented_targets,
                            raw_target_audios=raw_target_audios)

                        for snr in mixdb.snrs:
                            mixdb.mixtures.append(MRecord(
                                target_file_index=target_file_index,
                                target_augmentation_index=target_augmentation_index,
                                noise_file_index=noise_file_index,
                                noise_offset=noise_offset,
                                noise_augmentation_index=noise_augmentation_index,
                                snr=snr,
                                random_snr=False,
                            ))

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length

                        for random_snr in mixdb.random_snrs:
                            mixdb.mixtures.append(MRecord(
                                target_file_index=target_file_index,
                                target_augmentation_index=target_augmentation_index,
                                noise_file_index=noise_file_index,
                                noise_offset=noise_offset,
                                noise_augmentation_index=noise_augmentation_index,
                                snr=evaluate_random_rule(random_snr),
                                random_snr=True,
                            ))

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length

    else:
        # Get indices and offsets
        noise_offset = 0
        noise_file_index = 0
        noise_augmentation_index = 0
        for mixup in mixups:
            augmented_target_indices_for_mixup = get_augmented_target_indices_for_mixup(
                mixdb=mixdb,
                augmented_targets=augmented_targets,
                mixup=mixup)
            for augmented_target_index_for_mixup in augmented_target_indices_for_mixup:
                (target_file_index,
                 target_augmentation_index,
                 target_length) = _get_target_info(
                    mixdb=mixdb,
                    augmented_target_indices_for_mixup=augmented_target_index_for_mixup,
                    augmented_targets=augmented_targets,
                    raw_target_audios=raw_target_audios)

                for snr in mixdb.snrs:
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_offset) = _get_next_noise_offset(target_length=target_length,
                                                            augmented_noise_audios=augmented_noise_audios,
                                                            noise_file_index=noise_file_index,
                                                            noise_augmentation_index=noise_augmentation_index,
                                                            noise_offset=noise_offset)

                    mixdb.mixtures.append(MRecord(target_file_index=target_file_index,
                                                  target_augmentation_index=target_augmentation_index,
                                                  noise_file_index=noise_file_index,
                                                  noise_augmentation_index=noise_augmentation_index,
                                                  noise_offset=noise_offset,
                                                  snr=snr,
                                                  random_snr=False,
                                                  ))

                    noise_offset += target_length
                    used_noise_samples += target_length

                for random_snr in mixdb.random_snrs:
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_offset) = _get_next_noise_offset(target_length=target_length,
                                                            augmented_noise_audios=augmented_noise_audios,
                                                            noise_file_index=noise_file_index,
                                                            noise_augmentation_index=noise_augmentation_index,
                                                            noise_offset=noise_offset)

                    mixdb.mixtures.append(MRecord(target_file_index=target_file_index,
                                                  target_augmentation_index=target_augmentation_index,
                                                  noise_file_index=noise_file_index,
                                                  noise_augmentation_index=noise_augmentation_index,
                                                  noise_offset=noise_offset,
                                                  snr=evaluate_random_rule(random_snr),
                                                  random_snr=True,
                                                  ))

                    noise_offset += target_length
                    used_noise_samples += target_length

    # Fill in the details
    MP_DICT['mixdb'] = mixdb
    MP_DICT['raw_target_audios'] = raw_target_audios
    MP_DICT['augmented_noise_audios'] = augmented_noise_audios
    MP_DICT['save_mix'] = save_mix
    MP_DICT['save_ft'] = save_ft
    MP_DICT['save_segsnr'] = save_segsnr
    MP_DICT['output_dir'] = output_dir

    progress = tqdm(total=total_mixtures, desc='genmixdb', disable=not show_progress)
    mixdb.mixtures = p_tqdm_map(_process_mixture, range(total_mixtures), progress=progress)
    progress.close()

    mixdb.class_count = get_class_count_from_mixids(mixdb)

    total_samples = sum([sub.samples for sub in mixdb.mixtures])
    total_duration = total_samples / sonusai.mixture.SAMPLE_RATE
    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb.num_classes,
                               feature_step_samples=fs.feature_step_samples,
                               num_bands=fs.num_bands,
                               stride=fs.stride,
                               desc='Actual')
        noise_samples_percent = (float(used_noise_samples) / float(noise_audio_samples)) * 100
        logger.info('')
        logger.info(f'Used {noise_samples_percent:,.0f}% of noise audio')
        logger.info('')

    return mixdb


def _get_next_noise_offset(target_length: int,
                           augmented_noise_audios: List[List[np.ndarray]],
                           noise_file_index: int,
                           noise_augmentation_index: int,
                           noise_offset: int) -> (int, int, int):
    from sonusai import SonusAIError

    if noise_offset + target_length >= len(augmented_noise_audios[noise_file_index][noise_augmentation_index]):
        if noise_offset == 0:
            raise SonusAIError('Length of target audio exceeds length of noise audio')

        noise_offset = 0
        noise_augmentation_index += 1
        if noise_augmentation_index == len(augmented_noise_audios[noise_file_index]):
            noise_augmentation_index = 0
            noise_file_index += 1
            if noise_file_index == len(augmented_noise_audios):
                noise_file_index = 0

    return noise_file_index, noise_augmentation_index, noise_offset


def _get_target_info(mixdb: MixtureDatabase,
                     augmented_target_indices_for_mixup: List[int],
                     augmented_targets: AugmentedTargets,
                     raw_target_audios: List[np.ndarray]) -> (List[int], List[int], int):
    from sonusai.mixture import estimate_audio_length

    target_file_index = []
    target_augmentation_index = []
    target_length = 0
    for idx in augmented_target_indices_for_mixup:
        tfi = augmented_targets[idx].target_file_index
        tai = augmented_targets[idx].target_augmentation_index

        target_file_index.append(tfi)
        target_augmentation_index.append(tai)

        raw_target_audio = raw_target_audios[tfi]
        target_augmentation = mixdb.target_augmentations[tai]
        target_length = max(estimate_audio_length(audio_in=raw_target_audio,
                                                  augmentation=target_augmentation,
                                                  length_common_denominator=mixdb.feature_step_samples),
                            target_length)
    return target_file_index, target_augmentation_index, target_length


def get_output_from_config(config: dict, config_name: str) -> str:
    from os.path import splitext

    from sonusai import SonusAIError

    try:
        return config['output'].replace('${config}', splitext(config_name)[0])
    except Exception as e:
        raise SonusAIError(f'Error getting genmixdb base name: {e}')


def _process_mixture(mixid: int) -> MRecord:
    from os.path import join

    import h5py

    from sonusai.genft import genft
    from sonusai.mixture import get_mixid_padded_name
    from sonusai.mixture import get_mixture_metadata
    from sonusai.mixture import process_mixture

    mrecord = MP_DICT['mixdb'].mixtures[mixid]

    (mrecord,
     mixture,
     target,
     noise,
     truth_t) = process_mixture(mixdb=MP_DICT['mixdb'],
                                mrecord=mrecord,
                                raw_target_audios=MP_DICT['raw_target_audios'],
                                augmented_noise_audios=MP_DICT['augmented_noise_audios'])

    if MP_DICT['save_mix'] or MP_DICT['save_ft']:
        mixid_padded_name = get_mixid_padded_name(MP_DICT['mixdb'], mixid)
        output_base = join(MP_DICT['output_dir'], mixid_padded_name)

        if MP_DICT['save_mix']:
            with h5py.File(output_base + '.h5', 'a') as f:
                datasets = ['mixture', 'target', 'noise']
                for dataset in datasets:
                    if dataset in f:
                        del f[dataset]

                f.create_dataset(name='mixture', data=mixture, dtype=np.int16)
                f.create_dataset(name='target', data=target, dtype=np.int16)
                f.create_dataset(name='noise', data=noise, dtype=np.int16)

        if MP_DICT['save_ft']:
            feature, truth_f, segsnr = genft(mixdb=MP_DICT['mixdb'],
                                             mixids=mixid,
                                             raw_target_audios=MP_DICT['raw_target_audios'],
                                             augmented_noise_audios=MP_DICT['augmented_noise_audios'],
                                             mixture=mixture,
                                             truth_t=truth_t,
                                             compute_segsnr=MP_DICT['save_segsnr'])

            with h5py.File(output_base + '.h5', 'a') as f:
                datasets = ['feature', 'truth_f', 'segsnr']
                for dataset in datasets:
                    if dataset in f:
                        del f[dataset]

                f.create_dataset(name='feature', data=feature, dtype=np.single)
                f.create_dataset(name='truth_f', data=truth_f, dtype=np.single)
                if MP_DICT['save_segsnr']:
                    f.create_dataset(name='segsnr', data=segsnr, dtype=np.single)

        with open(file=output_base + '.txt', mode='w') as f:
            f.write(get_mixture_metadata(MP_DICT['mixdb'], MP_DICT['mixdb'].mixtures[mixid]))

    return mrecord


def main():
    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir

    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import load_config
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    save_mix = args['--mix']
    save_ft = args['--ft']
    save_segsnr = args['--segsnr']

    for config_file in args['CONFIG']:
        start_time = time.monotonic()
        logger.info(f'Creating mixture database for {config_file}')
        config = load_config(config_file)
        output = get_output_from_config(config, config_file)

        if save_mix or save_ft:
            if exists(output) and not isdir(output):
                remove(output)

            makedirs(output, exist_ok=True)

        create_file_handler(output + '.log')
        update_console_handler(verbose)
        initial_log_messages('genmixdb')

        mixdb = genmixdb(config=config,
                         save_mix=save_mix,
                         save_ft=save_ft,
                         save_segsnr=save_segsnr,
                         output_dir=output,
                         show_progress=True)

        json_name = output + '.json'
        with open(file=json_name, mode='w') as file:
            file.write(mixdb.to_json(indent=2))
            logger.info(f'Wrote mixture database for {config_file} to {json_name}')

        end_time = time.monotonic()
        logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from sonusai.mixture.dataclasses_sonusai import DataClassSonusAIMixin
from sonusai.mixture.segment import Segment


@dataclass(frozen=True)
class TruthSetting(DataClassSonusAIMixin):
    index: Optional[List[int]] = None
    function: Optional[str] = None
    config: Optional[dict] = None


TruthSettings = List[TruthSetting]

OptionalNumberStr = Optional[Union[float, int, str]]
OptionalListNumberStr = Optional[List[Union[float, int, str]]]


@dataclass
class Augmentation(DataClassSonusAIMixin):
    normalize: OptionalNumberStr = None
    pitch: OptionalNumberStr = None
    tempo: OptionalNumberStr = None
    gain: OptionalNumberStr = None
    eq1: OptionalListNumberStr = None
    eq2: OptionalListNumberStr = None
    eq3: OptionalListNumberStr = None
    lpf: OptionalNumberStr = None
    count: Optional[int] = None
    mixup: Optional[int] = 1


Augmentations = List[Augmentation]


@dataclass(frozen=True)
class TargetFile(DataClassSonusAIMixin):
    name: str
    duration: float
    truth_settings: TruthSettings
    augmentations: Optional[Augmentations] = None
    class_balancing_augmentation: Optional[Augmentation] = None


TargetFiles = List[TargetFile]


@dataclass
class AugmentedTarget(DataClassSonusAIMixin):
    target_file_index: int
    target_augmentation_index: int


AugmentedTargets = List[AugmentedTarget]


@dataclass(frozen=True)
class NoiseFile(DataClassSonusAIMixin):
    name: str
    duration: float
    augmentations: Optional[Augmentations] = None


NoiseFiles = List[NoiseFile]

ClassCount = List[int]


@dataclass
class MRecord(DataClassSonusAIMixin):
    target_file_index: List[int] = None
    noise_file_index: int = None
    noise_offset: int = None
    target_augmentation_index: List[int] = None
    noise_augmentation_index: int = None
    snr: float = None
    random_snr: Optional[bool] = None
    samples: int = None
    target_gain: List[int] = None
    class_count: ClassCount = None
    target_snr_gain: float = None
    noise_snr_gain: float = None
    o_frame_offset: Optional[int] = None


MRecords = List[MRecord]

MixtureIDs = Union[str, int, List[int], range]


@dataclass
class MixtureDatabase(DataClassSonusAIMixin):
    class_balancing: Optional[bool] = False
    class_balancing_augmentation: Optional[Augmentation] = None
    class_count: ClassCount = None
    class_labels: List[str] = None
    class_weights_threshold: List[float] = None
    exhaustive_noise: Optional[bool] = True
    feature: str = None
    feature_samples: int = None
    feature_step_samples: int = None
    first_cba_index: Optional[int] = None
    frame_size: int = None
    mixid_padding: Optional[int] = None
    mixtures: MRecords = None
    noise_augmentations: Augmentations = None
    noises: NoiseFiles = None
    num_classes: int = None
    random_snrs: Optional[List[str]] = None
    seed: Optional[int] = 0
    snrs: List[float] = None
    target_augmentations: Augmentations = None
    targets: TargetFiles = None
    truth_mutex: bool = None
    truth_reduction_function: str = None
    truth_settings: TruthSettings = None


def convert_json_to_mixdb(data: str) -> MixtureDatabase:
    import json

    return MixtureDatabase.from_dict(json.loads(data))


def load_mixdb(name: str) -> MixtureDatabase:
    from os.path import exists
    from os.path import splitext

    import h5py

    from sonusai import SonusAIError

    if not exists(name):
        raise SonusAIError(f'{name} does not exist')

    ext = splitext(name)[1]

    if ext == '.json':
        with open(file=name, mode='r', encoding='utf-8') as f:
            json_mixdb = f.read()
    elif ext == '.h5':
        with h5py.File(name, 'r') as f:
            json_mixdb = f.attrs['mixdb']
    else:
        raise SonusAIError(f'Do not know how to load mixdb from {name}')

    return convert_json_to_mixdb(json_mixdb)


def get_samples_in_mixid(mixdb: MixtureDatabase, mixid: int) -> int:
    return mixdb.mixtures[mixid].samples


def get_transform_frames_in_mrecord(mixdb: MixtureDatabase, mrecord: MRecord) -> int:
    return mrecord.samples // mixdb.frame_size


def get_feature_frames_in_mrecord(mixdb: MixtureDatabase, mrecord: MRecord) -> int:
    return mrecord.samples // mixdb.feature_step_samples


def get_class_weights_threshold(mixdb: Union[MixtureDatabase, Dict]) -> List[float]:
    """Get the class_weights_threshold from a mixture database or a config."""

    from sonusai import SonusAIError

    if isinstance(mixdb, dict):
        class_weights_threshold = mixdb['class_weights_threshold']
        num_classes = mixdb['num_classes']
    else:
        class_weights_threshold = mixdb.class_weights_threshold
        num_classes = mixdb.num_classes

    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold] * num_classes

    if len(class_weights_threshold) != num_classes:
        raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')

    return class_weights_threshold


def get_file_frame_segments(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> Dict[int, Segment]:
    """
    Get Segment data from a mixture database
    Return a dictionary where:
        - keys are the mixids
        - values are file frame Segment objects
    """
    _mixid = convert_mixids_to_list(mixdb, mixids)

    # remove duplicates
    _mixid = sorted(list(set(_mixid)))

    file_frame_segments = {}
    for m in _mixid:
        file_frame_segments[m] = Segment(mixdb.mixtures[m].o_frame_offset,
                                         get_feature_frames_in_mrecord(mixdb, mixdb.mixtures[m]))

    return file_frame_segments


def new_mixdb_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs) -> MixtureDatabase:
    from copy import deepcopy

    from sonusai import SonusAIError

    mixdb_out = deepcopy(mixdb)
    mixdb_out.mixtures = get_mrecords_from_mixids(mixdb_out, mixids)

    if not mixdb_out.mixtures:
        raise SonusAIError(f'Error processing mixid: {mixids}; resulted in empty list of mixtures')

    return mixdb_out


def convert_mixids_to_list(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> List[int]:
    from sonusai import SonusAIError

    result = mixids
    all_mixids = list(range(len(mixdb.mixtures)))

    if result is None:
        result = all_mixids

    if isinstance(result, str):
        if result == '*':
            result = all_mixids
        else:
            try:
                result = eval(f'{all_mixids}[{result}]')
            except NameError:
                result = []

    if isinstance(result, range):
        result = list(result)

    if isinstance(result, int):
        result = [result]

    if not all(isinstance(x, int) and 0 <= x < len(mixdb.mixtures) for x in result):
        raise SonusAIError(f'Invalid entries in mixids of {mixids}')

    if not result:
        raise SonusAIError(f'Empty mixids {mixids}')

    return result


def get_mrecords_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> MRecords:
    from copy import deepcopy

    return [deepcopy(mixdb.mixtures[i]) for i in convert_mixids_to_list(mixdb, mixids)]


def get_mixids_data(mixdb: MixtureDatabase,
                    mixids: MixtureIDs,
                    truth_f: np.ndarray,
                    predict: np.ndarray) -> (np.ndarray, np.ndarray):
    """Collect per-feature data of specified mixids from mixdb where inputs are:
       truth_f:   truth data matching mixdb (size feature_frames x num_classes)
       predict:   prediction or segsnr data (size feature_frames x ndim, where ndim > 1)

    Returns:
        truth_f_out:    np.ndarray combined truth from mixids
        predict_out:    np.ndarray combined data from mixids
    """
    from sonusai import SonusAIError
    from sonusai.mixture import get_slices

    in_frames = truth_f.shape[0]

    if in_frames != predict.shape[0]:
        raise SonusAIError('truth_f and predict have different number of frames')

    num_classes = truth_f.shape[1]
    # same as num_class for prediction data, but use for segsnr too
    num_data = predict.shape[1]

    _mixid = convert_mixids_to_list(mixdb, mixids)
    # TODO: Eliminate file_frame_segments
    file_frame_segments = get_file_frame_segments(mixdb, _mixid)
    out_frames = sum([file_frame_segments[m].length for m in _mixid])
    truth_f_out = np.empty((out_frames, num_classes), dtype=np.single)
    predict_out = np.empty((out_frames, num_data), dtype=np.single)

    # Handle the special case when input data is smaller, i.e., prediction when total mixture
    # length is a non-multiple of the batch size. In this case just pad both with zeros; should
    # have negligible effect on metrics
    max_in_index = max([file_frame_segments[m].get_slice().stop for m in _mixid])
    frame_diff = max_in_index - in_frames
    if frame_diff > 0:
        truth_f = np.concatenate((truth_f, np.zeros((frame_diff, num_classes))))
        predict = np.concatenate((predict, np.zeros((frame_diff, num_classes))))

    start = 0
    for m in _mixid:
        in_slice, out_slice = get_slices(file_frame_segments[m], start)

        if len(truth_f[in_slice]) != len(truth_f_out[out_slice]):
            raise SonusAIError(f'in_slice and out_slice are different lengths')

        truth_f_out[out_slice] = truth_f[in_slice]
        predict_out[out_slice] = predict[in_slice]
        start += file_frame_segments[m].length

    return truth_f_out, predict_out


def evaluate_random_rule(rule: str) -> Union[str, float]:
    """Evaluate 'rand' directive."""
    import re
    from random import uniform

    from sonusai.mixture import RAND_PATTERN

    def rand_repl(m):
        return f'{uniform(float(m.group(1)), float(m.group(4))):.2f}'

    return eval(re.sub(RAND_PATTERN, rand_repl, rule))


def get_mixid_padded_name(mixdb: MixtureDatabase, mixid: int) -> str:
    return f'{mixid:0{mixdb.mixid_padding}}'


def get_mixture_metadata(mixdb: MixtureDatabase, mrecord: MRecord) -> str:
    metadata = ''
    for ti in range(len(mrecord.target_file_index)):
        metadata += f'target {ti} name: {mixdb.targets[ti].name}\n'
        metadata += f'target {ti} augmentation: {mixdb.target_augmentations[ti].to_dict()}\n'
        metadata += f'target {ti} target_gain: {mrecord.target_gain[ti]}\n'
        ts = mixdb.targets[ti].truth_settings
        for tsi in range(len(ts)):
            metadata += f'target {ti} truth index {tsi}: {ts[tsi].index}\n'
    metadata += f'noise name: {mixdb.noises[mrecord.noise_file_index].name}\n'
    metadata += f'noise augmentation: {mixdb.noise_augmentations[mrecord.noise_augmentation_index].to_dict()}\n'
    metadata += f'snr: {mrecord.snr}\n'
    metadata += f'random_snr: {mrecord.random_snr}\n'
    metadata += f'samples: {mrecord.samples}\n'
    metadata += f'target_snr_gain: {mrecord.target_snr_gain}\n'
    metadata += f'noise_snr_gain: {mrecord.noise_snr_gain}\n'
    metadata += f'class_count: {mrecord.class_count}\n'

    return metadata



from dataclasses import dataclass
from typing import List

import numpy as np
from tensorflow.keras.utils import Sequence


class SequenceFromH5(Sequence):
    """Generates data from a directory containing genft data"""

    @dataclass(frozen=True)
    class BatchParams:
        idx: List[int]
        offset: int
        extra: int

    def __init__(self,
                 files: List[str],
                 feature: str,
                 num_classes: int,
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 truth_mutex: bool):
        """Initialization"""
        from os.path import exists

        import h5py

        from sonusai import SonusAIError
        from sonusai.data_generator import get_frames_per_batch
        from sonusai.mixture import get_feature_stats

        self.files = files
        self.feature = feature
        self.num_classes = num_classes
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.truth_mutex = truth_mutex

        for file in self.files:
            if not exists(file):
                raise SonusAIError(f'Could not find {file}')
            with h5py.File(file, 'r') as f:
                if 'feature' not in f:
                    raise SonusAIError(f'Could not find "feature" in {file}')

        fs = get_feature_stats(feature=self.feature, num_classes=self.num_classes)

        self.stride = fs.stride
        self.num_bands = fs.num_bands
        self.frames_per_batch = get_frames_per_batch(self.batch_size, self.timesteps)

        # Compute idx, offset, and extra for dataset
        # offsets and extras are needed because files are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of files, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_file_index = 0
        offset = 0
        self.batch_params = []
        self.file_indices = []
        total_frames = 0
        for i in range(len(self.files)):
            current_frames = self._get_feature(i).shape[0]
            self.file_indices.append(slice(total_frames, total_frames + current_frames))
            total_frames += current_frames
            cumulative_frames += current_frames
            while cumulative_frames >= self.frames_per_batch:
                extra_frames = cumulative_frames - self.frames_per_batch
                idx = list(range(start_file_index, i + 1))
                self.batch_params.append(self.BatchParams(idx=idx, offset=offset, extra=extra_frames))
                if extra_frames == 0:
                    start_file_index = i + 1
                    offset = 0
                else:
                    start_file_index = i
                    offset = current_frames - extra_frames
                cumulative_frames = extra_frames

        if cumulative_frames > 0:
            extra_frames = cumulative_frames - self.frames_per_batch
            idx = list(range(start_file_index, len(self.files)))
            self.batch_params.append(self.BatchParams(idx=idx, offset=offset, extra=extra_frames))

        self.total_batches = int(np.ceil(total_frames / self.frames_per_batch))

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> np.ndarray:
        """Get one batch of data"""
        from sonusai.utils import reshape_inputs

        batch_params = self.batch_params[batch_index]

        result = []
        for idx in batch_params.idx:
            result.append(self._get_feature(idx))

        feature = np.vstack([result[i] for i in range(len(result))])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            if batch_params.extra < 0:
                feature = np.pad(array=feature,
                                 pad_width=((0, -batch_params.extra), (0, 0), (0, 0)),
                                 mode='constant',
                                 constant_values=0)

        feature, _, _, _, _, _ = reshape_inputs(feature=feature,
                                                batch_size=self.batch_size,
                                                timesteps=self.timesteps,
                                                flatten=self.flatten,
                                                add1ch=self.add1ch)

        return feature

    def _get_feature(self, idx: int) -> np.ndarray:
        import h5py

        with h5py.File(self.files[idx], 'r') as f:
            return np.array(f['feature'])



# SonusAI data generator classes

from sonusai.data_generator.keras_from_h5 import KerasFromH5
from sonusai.data_generator.keras_from_mixdb import KerasFromMixDB
from sonusai.data_generator.sequence_from_h5 import SequenceFromH5
from sonusai.data_generator.utils import get_frames_per_batch



from os import PathLike
from pathlib import Path
from typing import Dict
from typing import List
from typing import Union

from sonusai.mixture.mixdb import NoiseFiles
from sonusai.mixture.mixdb import TargetFiles


def raw_load_config(name: str) -> Dict:
    """Load YAML file with SonusAI variable substitution."""
    import yaml

    with open(file=name, mode='r') as f:
        config = yaml.safe_load(f)

    return config_variable_substitution(config)


def config_variable_substitution(config: Dict) -> Dict:
    """Find custom SonusAI variables in given dictionary and substitute their values in place."""
    import json

    from sonusai.mixture import DEFAULT_FRAME_SIZE

    result = json.dumps(config)
    result = result.replace('"${frame_size}"', str(DEFAULT_FRAME_SIZE))
    config = json.loads(result)
    return config


def get_default_config() -> Dict:
    """Load default SonusAI config."""
    from sonusai import SonusAIError
    from sonusai.mixture import DEFAULT_CONFIG

    try:
        return raw_load_config(DEFAULT_CONFIG)
    except Exception as e:
        raise SonusAIError(f'Error loading default config: {e}')


def load_config(name: str) -> Dict:
    """Load SonusAI default config and update with given YAML file (performing SonusAI variable substitution)."""
    return update_config_from_file(name=name, config=get_default_config())


def update_config_from_file(name: Union[str, bytes, PathLike], config: Dict) -> Dict:
    """Update the given config with the config in the given YAML file."""
    from copy import deepcopy

    from sonusai import SonusAIError
    from sonusai.mixture import REQUIRED_CONFIGS
    from sonusai.mixture import VALID_CONFIGS

    new_config = deepcopy(config)

    try:
        given_config = raw_load_config(name)
    except Exception as e:
        raise SonusAIError(f'Error loading config from {name}: {e}')

    # Check for unrecognized keys
    for key in given_config:
        if key not in VALID_CONFIGS:
            nice_list = '\n'.join([f'  {item}' for item in VALID_CONFIGS])
            raise SonusAIError(f'Invalid config parameter in {name}: {key}.\nValid config parameters are:\n{nice_list}')

    # Use default config as base and overwrite with given config keys as found
    for key in new_config:
        if key in given_config:
            if key not in ['truth_settings']:
                new_config[key] = given_config[key]

    # Handle 'truth_settings' special case
    if 'truth_settings' in given_config:
        new_config['truth_settings'] = deepcopy(given_config['truth_settings'])

    if not isinstance(new_config['truth_settings'], list):
        new_config['truth_settings'] = [new_config['truth_settings']]

    default = deepcopy(config['truth_settings'])
    if not isinstance(default, list):
        default = [default]

    new_config['truth_settings'] = update_truth_settings(new_config['truth_settings'], default)

    # Check for required keys
    for key in REQUIRED_CONFIGS:
        if key not in new_config:
            raise SonusAIError(f'Missing required config in {name}: {key}')

    return new_config


def update_truth_settings(given: Union[List[Dict], Dict], default: List[Dict] = None) -> List[Dict]:
    """Update missing fields in given 'truth_settings' with default values."""
    from copy import deepcopy

    from sonusai import SonusAIError
    from sonusai.mixture import VALID_TRUTH_SETTINGS

    truth_settings = deepcopy(given)
    if not isinstance(truth_settings, list):
        truth_settings = [truth_settings]

    if default is not None and len(truth_settings) != len(default):
        raise SonusAIError(f'Length of given does not match default')

    for n in range(len(truth_settings)):
        for key in truth_settings[n]:
            if key not in VALID_TRUTH_SETTINGS:
                nice_list = '\n'.join([f'  {item}' for item in VALID_TRUTH_SETTINGS])
                raise SonusAIError(f'Invalid truth_settings: {key}.\nValid truth_settings are:\n{nice_list}')

        for key in VALID_TRUTH_SETTINGS:
            if key not in truth_settings[n]:
                if default is not None and key in default[n]:
                    truth_settings[n][key] = default[n][key]
                else:
                    raise SonusAIError(f'Missing required truth_settings: {key}')

    for truth_setting in truth_settings:
        if not isinstance(truth_setting['index'], list):
            truth_setting['index'] = [truth_setting['index']]

    return truth_settings


def get_hierarchical_config_files(root: str, leaf: str) -> List[Path]:
    """Get a hierarchical list of config files in the given leaf of the given root."""
    from sonusai import SonusAIError
    import os

    config_file = 'config.yml'

    root_path = Path(os.path.abspath(root))
    if not root_path.is_dir():
        raise SonusAIError(f'Given root, {root_path}, is not a directory.')

    leaf_path = Path(os.path.abspath(leaf))
    if not leaf_path.is_dir():
        raise SonusAIError(f'Given leaf, {leaf_path}, is not a directory.')

    common = os.path.commonpath((root_path, leaf_path))
    if os.path.normpath(common) != os.path.normpath(root_path):
        raise SonusAIError(f'Given leaf, {leaf_path}, is not in the hierarchy of the given root, {root_path}')

    top_config_file = Path(os.path.join(root_path, config_file))
    if not top_config_file.is_file():
        raise SonusAIError(f'Could not find {top_config_file}')

    current = leaf_path
    config_files = []
    while current != root_path:
        local_config_file = Path(os.path.join(current, config_file))
        if local_config_file.is_file():
            config_files.append(local_config_file)
        current = current.parent

    config_files.append(top_config_file)
    return list(reversed(config_files))


def update_config_from_hierarchy(root: str, leaf: str, config: dict) -> dict:
    """Update the given config using the hierarchical config files in the given leaf of the given root."""
    from copy import deepcopy

    new_config = deepcopy(config)
    config_files = get_hierarchical_config_files(root=root, leaf=leaf)
    for config_file in config_files:
        new_config = update_config_from_file(name=config_file, config=new_config)

    return new_config


def get_max_class(num_classes: int, truth_mutex: bool) -> int:
    """Get the maximum class index."""
    max_class = num_classes
    if truth_mutex:
        max_class -= 1
    return max_class


def get_target_files(config: dict) -> TargetFiles:
    from sonusai import SonusAIError
    from sonusai.utils import dataclass_from_dict

    truth_settings = config.get('truth_settings', list())
    target_files = []
    for target_file in config['targets']:
        append_target_files(target_files, target_file, truth_settings)

    max_class = get_max_class(config['num_classes'], config['truth_mode'] == 'mutex')

    for target_file in target_files:
        target_file['truth_settings'] = update_truth_settings(target_file['truth_settings'], config['truth_settings'])

        for truth_setting in target_file['truth_settings']:
            if any(idx > max_class for idx in truth_setting['index']):
                raise SonusAIError('invalid truth index')

    return dataclass_from_dict(TargetFiles, target_files)


def append_target_files(target_files: List[dict],
                        target_file: Union[dict, str],
                        truth_settings: List[dict],
                        tokens: Union[dict, None] = None) -> None:
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import sox

    from sonusai import SonusAIError

    if tokens is None:
        tokens = {}

    if isinstance(target_file, dict):
        if 'name' in target_file:
            in_name = target_file['name']
        else:
            raise SonusAIError('Target list contained record without name')

        if 'truth_settings' in target_file:
            truth_settings = target_file['truth_settings']
    else:
        in_name = target_file

    in_name, new_tokens = tokenized_expandvars(in_name)
    tokens.update(new_tokens)
    names = glob(in_name)
    if not names:
        raise SonusAIError(f'Could not find {in_name}. Make sure path exists')
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                append_target_files(target_files, child, truth_settings, tokens)
        else:
            try:
                if ext == '.txt':
                    with open(file=name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expandvars(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                append_target_files(target_files, child, truth_settings, tokens)
                elif ext == '.yml':
                    try:
                        yml_config = raw_load_config(name)

                        if 'targets' in yml_config:
                            for record in yml_config['targets']:
                                append_target_files(target_files, record, truth_settings, tokens)
                    except Exception as e:
                        raise SonusAIError(f'Error processing {name}: {e}')
                else:
                    sox.file_info.validate_input_file(name)
                    duration = sox.file_info.duration(name)
                    for key, value in tokens.items():
                        name = name.replace(value, f'${key}')
                    entry = {
                        'name':     name,
                        'duration': duration,
                    }
                    if len(truth_settings) > 0:
                        entry['truth_settings'] = truth_settings
                        for truth_setting in entry['truth_settings']:
                            if 'function' in truth_setting and truth_setting['function'] == 'file':
                                truth_setting['config']['file'] = splitext(name)[0] + '.h5'
                    target_files.append(entry)
            except SonusAIError:
                raise
            except Exception as e:
                raise SonusAIError(f'Error processing {name}: {e}')


def get_noise_files(config: dict) -> NoiseFiles:
    from sonusai.utils import dataclass_from_dict

    noise_files = []
    for noise_file in config['noises']:
        append_noise_files(noise_files, noise_file)

    return dataclass_from_dict(NoiseFiles, noise_files)


def append_noise_files(noise_files: List[dict],
                       noise_file: Union[dict, str],
                       tokens: Union[dict, None] = None) -> None:
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import sox

    from sonusai import SonusAIError

    if tokens is None:
        tokens = {}

    if isinstance(noise_file, dict):
        if 'name' in noise_file:
            in_name = noise_file['name']
        else:
            raise SonusAIError('Noise list contained record without name')
    else:
        in_name = noise_file

    in_name, new_tokens = tokenized_expandvars(in_name)
    tokens.update(new_tokens)
    names = glob(in_name)
    if not names:
        raise SonusAIError(f'Could not find {in_name}. Make sure path exists')
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                append_noise_files(noise_files, child, tokens)
        else:
            try:
                if ext == '.txt':
                    with open(file=name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expandvars(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                append_noise_files(noise_files, child, tokens)
                elif ext == '.yml':
                    try:
                        yml_config = raw_load_config(name)

                        if 'noises' in yml_config:
                            for record in yml_config['noises']:
                                append_noise_files(noise_files, record, tokens)
                    except Exception as e:
                        raise SonusAIError(f'Error processing {name}: {e}')
                else:
                    sox.file_info.validate_input_file(name)
                    duration = sox.file_info.duration(name)
                    for key, value in tokens.items():
                        name = name.replace(value, f'${key}')
                    entry = {
                        'name':     name,
                        'duration': duration,
                    }
                    noise_files.append(entry)
            except SonusAIError:
                raise
            except Exception as e:
                raise SonusAIError(f'Error processing {name}: {e}')


def tokenized_expandvars(path: str) -> (str, dict):
    """Expand shell variables of the forms $var, ${var} and %var%.
    Unknown variables are left unchanged."""

    # Expand paths containing shell variable substitutions.
    # The following rules apply:
    #       - no expansion within single quotes
    #       - '$$' is translated into '$'
    #       - '%%' is translated into '%' if '%%' are not seen in %var1%%var2%
    #       - ${varname} is accepted.
    #       - $varname is accepted.
    #       - %varname% is accepted.
    #       - varnames can be made out of letters, digits and the characters '_-'
    #         (though is not verified in the ${varname} and %varname% cases)
    # XXX With COMMAND.COM you can use any characters in a variable name,
    # XXX except '^|<>='.
    import os
    import string

    from sonusai.mixture import DEFAULT_NOISE

    os.environ['default_noise'] = DEFAULT_NOISE

    path = os.fspath(path)
    token_map = {}
    if isinstance(path, bytes):
        if b'$' not in path and b'%' not in path:
            return path, token_map
        varchars = bytes(string.ascii_letters + string.digits + '_-', 'ascii')
        quote = b'\''
        percent = b'%'
        brace = b'{'
        rbrace = b'}'
        dollar = b'$'
        environ = getattr(os, 'environb', None)
    else:
        if '$' not in path and '%' not in path:
            return path, token_map
        varchars = string.ascii_letters + string.digits + '_-'
        quote = '\''
        percent = '%'
        brace = '{'
        rbrace = '}'
        dollar = '$'
        environ = os.environ
    res = path[:0]
    index = 0
    pathlen = len(path)
    while index < pathlen:
        c = path[index:index + 1]
        if c == quote:  # no expansion within single quotes
            path = path[index + 1:]
            pathlen = len(path)
            try:
                index = path.index(c)
                res += c + path[:index + 1]
            except ValueError:
                res += c + path
                index = pathlen - 1
        elif c == percent:  # variable or '%'
            if path[index + 1:index + 2] == percent:
                res += c
                index += 1
            else:
                path = path[index + 1:]
                pathlen = len(path)
                try:
                    index = path.index(percent)
                except ValueError:
                    res += percent + path
                    index = pathlen - 1
                else:
                    var = path[:index]
                    try:
                        if environ is None:
                            value = os.fsencode(os.environ[os.fsdecode(var)])
                        else:
                            value = environ[var]
                        token_map[var] = value
                    except KeyError:
                        value = percent + var + percent
                    res += value
        elif c == dollar:  # variable or '$$'
            if path[index + 1:index + 2] == dollar:
                res += c
                index += 1
            elif path[index + 1:index + 2] == brace:
                path = path[index + 2:]
                pathlen = len(path)
                try:
                    index = path.index(rbrace)
                except ValueError:
                    res += dollar + brace + path
                    index = pathlen - 1
                else:
                    var = path[:index]
                    try:
                        if environ is None:
                            value = os.fsencode(os.environ[os.fsdecode(var)])
                        else:
                            value = environ[var]
                        token_map[var] = value
                    except KeyError:
                        value = dollar + brace + var + rbrace
                    res += value
            else:
                var = path[:0]
                index += 1
                c = path[index:index + 1]
                while c and c in varchars:
                    var += c
                    index += 1
                    c = path[index:index + 1]
                try:
                    if environ is None:
                        value = os.fsencode(os.environ[os.fsdecode(var)])
                    else:
                        value = environ[var]
                    token_map[var] = value
                except KeyError:
                    value = dollar + var
                res += value
                if c:
                    index -= 1
        else:
            res += c
        index += 1
    return res, token_map



import numpy as np


def calculate_snr_f_statistics(truth_f: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    """Calculate statistics of snr_f truth data.

    For now, includes mean and standard deviation of the raw values (usually energy)
    and mean and standard deviation of the dB values (10 * log10).
    """
    classes = truth_f.shape[1]

    snr_mean = np.zeros(classes, dtype=np.single)
    snr_std = np.zeros(classes, dtype=np.single)
    snr_db_mean = np.zeros(classes, dtype=np.single)
    snr_db_std = np.zeros(classes, dtype=np.single)

    for c in range(classes):
        tmp = truth_f[:, c]
        tmp = np.double(tmp[np.isfinite(tmp)])

        if len(tmp) == 0:
            snr_mean[c] = -np.inf
            snr_std[c] = -np.inf
        else:
            snr_mean[c] = np.mean(tmp)
            snr_std[c] = np.std(tmp, ddof=1)

        tmp2 = 10 * np.ma.log10(tmp).filled(-np.inf)
        tmp2 = tmp2[np.isfinite(tmp2)]

        if len(tmp2) == 0:
            snr_db_mean[c] = -np.inf
            snr_db_std[c] = -np.inf
        else:
            snr_db_mean[c] = np.mean(tmp2)
            snr_db_std[c] = np.std(tmp2, ddof=1)

    return snr_mean, snr_std, snr_db_mean, snr_db_std


def calculate_mapped_snr_f(truth_f: np.ndarray, snr_db_mean: np.ndarray, snr_db_std: np.ndarray) -> np.ndarray:
    """Calculate mapped SNR from standard SNR energy per bin/class."""
    import scipy.special as sc

    old_err = np.seterr(divide='ignore', invalid='ignore')
    num = 10 * np.log10(np.double(truth_f)) - np.double(snr_db_mean)
    den = np.double(snr_db_std) * np.sqrt(2)
    q = num / den
    q = np.nan_to_num(q, nan=-np.inf, posinf=np.inf, neginf=-np.inf)
    mapped_snr_f = 0.5 * (1 + sc.erf(q))
    np.seterr(**old_err)

    return np.single(mapped_snr_f)



from typing import List
from typing import Union

import numpy as np

from sonusai.mixture.mixdb import MixtureDatabase


def stratified_shuffle_split_mixid(mixdb: MixtureDatabase,
                                   vsplit: float = 0.2,
                                   nsplit: int = 0,
                                   rnd_seed: Union[int, None] = 0) -> (List[int], List[int], np.ndarray, np.ndarray):
    """
    Create a training and test/validation list of mixture IDs from all mixtures in a mixture database.
    The test/validation split is specified by vsplit (0.0 to 1.0), default 0.2.
    The mixtures are randomly shuffled by rnd_seed; set to int for repeatability, or None for no shuffle.
    The mixtures are then stratified across all populated classes.

    Inputs:
        mixdb:      Mixture database created by Aaware SonusAI genmixdb.
        vsplit:     Fractional split of mixtures for validation, 1-vsplit for training.
        nsplit:     Number of splits (TBD).
        rnd_seed:   Seed integer for reproducible random shuffling (or None for no shuffling).

    Outputs:
        t_mixid:        list of mixture IDs for training
        v_mixid:        list of mixture IDs for validation
        t_num_mixid:    list of class counts in t_mixid
        v_num_mixid:    list of class counts in v_mixid

    Examples:
        t_mixid, v_mixid, t_num_mixid, v_num_mixid = stratified_shuffle_split_mixid(mixdb, vsplit=vsplit)

    @author: Chris Eddington
    """
    import random
    from copy import deepcopy

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import get_truth_indices_for_target

    if vsplit < 0 or vsplit > 1:
        raise SonusAIError('vsplit must be between 0 and 1')

    a_class_mixid = {i + 1: [] for i in range(mixdb.num_classes)}
    for mixid, mixture in enumerate(mixdb.mixtures):
        if any(mixture.class_count) or mixdb.truth_mutex == 0:
            for truth_index in get_truth_indices_for_target(mixdb.targets[mixture.target_file_index[0]]):
                a_class_mixid[truth_index].append(mixid)
        else:
            # no counts and mutex mode means this is all 'other' class
            a_class_mixid[mixdb.num_classes].append(mixid)

    t_class_mixid = [[] for _ in range(mixdb.num_classes)]
    v_class_mixid = [[] for _ in range(mixdb.num_classes)]

    a_num_mixid = np.zeros(mixdb.num_classes, dtype=int)
    t_num_mixid = np.zeros(mixdb.num_classes, dtype=int)
    v_num_mixid = np.zeros(mixdb.num_classes, dtype=int)

    if rnd_seed is not None:
        random.seed(rnd_seed)

    # For each class pick percentage of shuffled mixids for training, validation
    for ci in range(mixdb.num_classes):
        # total number of mixids for class
        a_num_mixid[ci] = len(a_class_mixid[ci + 1])

        # number of training mixids for class
        t_num_mixid[ci] = int(np.floor(a_num_mixid[ci] * (1 - vsplit)))

        # number of validation mixids for class
        v_num_mixid[ci] = a_num_mixid[ci] - t_num_mixid[ci]

        # indices for all mixids in class
        indices = [*range(a_num_mixid[ci])]
        if rnd_seed is not None:
            # randomize order
            random.shuffle(indices)

        t_class_mixid[ci] = [a_class_mixid[ci + 1][ii] for ii in indices[0:t_num_mixid[ci]]]
        v_class_mixid[ci] = [a_class_mixid[ci + 1][ii] for ii in indices[t_num_mixid[ci]:]]

    if any(~(t_num_mixid > 0)):
        logger.warning(f'Some classes have zero coverage: {np.where(~(t_num_mixid > 0))[0]}')

    # Stratify over non-zero classes
    nz_indices = np.where(t_num_mixid > 0)[0]
    # First stratify pass is min count / 3 times through all classes, one each least populated class count (of non-zero)
    min_class = min(t_num_mixid[nz_indices])
    # number of mixids in each class for stratify by 1
    n0 = int(np.ceil(min_class / 3))
    # 3rd stage for stratify by 1
    n3 = int(n0)
    # 2nd stage stratify by class_count/min(class_count-n3) n2 times
    n2 = int(max(min_class - n0 - n3, 0))

    logger.info(f'Stratifying training, x1 cnt {n0}: x(class_count/{n2}): x1 cnt {n3} x1, '
                f'for {len(nz_indices)} populated classes')

    # initialize source list
    tt = deepcopy(t_class_mixid)
    t_num_mixid2 = deepcopy(t_num_mixid)
    t_mixid = []
    for _ in range(n0):
        for ci in range(mixdb.num_classes):
            if t_num_mixid2[ci] > 0:
                # append first
                t_mixid.append(tt[ci][0])
                del tt[ci][0]
                t_num_mixid2[ci] = len(tt[ci])

    # Now extract weighted by how many are left in class minus n3
    # which will leave approx n3 remaining
    if n2 > 0:
        # should always be non-zero
        min_class = np.min(t_num_mixid2 - n3)
        class_count = np.floor((t_num_mixid2 - n3) / min_class)
        # class_count = np.maximum(np.floor((t_num_mixid2 - n3) / n2),0) # Counts per class
        for _ in range(min_class):
            for ci in range(mixdb.num_classes):
                if class_count[ci] > 0:
                    for cc in range(int(class_count[ci])):
                        # append first
                        t_mixid.append(tt[ci][0])
                        del tt[ci][0]
                        t_num_mixid2[ci] = len(tt[ci])

    # Now extract remaining mixids, one each class until empty
    # There should be ~n3 remaining mixids in each
    _extract_remaining_mixids(mixdb, t_mixid, t_num_mixid2, tt)

    if len(t_mixid) != sum(t_num_mixid):
        logger.warning('Final stratified training list length does not match starting list length.')

    if any(t_num_mixid2) or any(tt):
        logger.warning('Remaining training mixid list not empty.')

    # Now stratify the validation list, which is probably not as important, so use simple method
    # initialize source list
    vv = deepcopy(v_class_mixid)
    v_num_mixid2 = deepcopy(v_num_mixid)
    v_mixid = []
    _extract_remaining_mixids(mixdb, v_mixid, v_num_mixid2, vv)

    if len(v_mixid) != sum(v_num_mixid):
        logger.warning('Final stratified validation list length does not match starting lists length.')

    if any(v_num_mixid2) or any(vv):
        logger.warning('Remaining validation mixid list not empty.')

    return t_mixid, v_mixid, t_num_mixid, v_num_mixid


def _extract_remaining_mixids(mixdb: MixtureDatabase,
                              mixid: List[int],
                              num_mixid: np.ndarray,
                              class_mixid: List[List[int]]) -> None:
    for _ in range(max(num_mixid)):
        for ci in range(mixdb.num_classes):
            if num_mixid[ci] > 0:
                # append first
                mixid.append(class_mixid[ci][0])
                del class_mixid[ci][0]
                num_mixid[ci] = len(class_mixid[ci])



"""Map functions with tqdm progress bars for parallel processing.

p_tqdm_map:  Performs a parallel ordered map.
p_tqdm_umap: Performs a parallel unordered map.
"""

from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import List


def _parallel(ordered: bool, function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a parallel map.

    Arguments:
        ordered     bool        True for an ordered map, False for an unordered map.
        function    Callable    The function to apply to each element of the given Iterable.
        iterables   Iterable    One or more Iterables containing the data to be mapped.

    Returns:
        A generator which will apply the function to each element of the given Iterables
        in parallel in order.
    """
    from typing import Sized

    from tqdm.auto import tqdm

    from sonusai.utils import initialize_map_func

    map_func = initialize_map_func(ordered, **kwargs)

    # Extract progress bar
    progress = kwargs.pop('progress', None)
    progress_needs_close = False

    if progress is None:
        # Determine length of tqdm (equal to length of the shortest iterable)
        total = kwargs.pop('total', None)
        if total is None:
            total = min(len(iterable) for iterable in iterables if isinstance(iterable, Sized))

        progress = tqdm(total=total, **kwargs)
        progress_needs_close = True

    for item in map_func(function, *iterables):
        yield item
        progress.update()

    if progress_needs_close:
        progress.close()


def p_tqdm_map(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel ordered map."""
    return list(_parallel(True, function, *iterables, **kwargs))


def p_tqdm_umap(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel unordered map."""
    return list(_parallel(False, function, *iterables, **kwargs))



# SonusAI mixture utilities
from sonusai.mixture.audio_db import check_audio_files_exist
from sonusai.mixture.audio_db import extract_results
from sonusai.mixture.audio_db import get_audio_and_truth_t
from sonusai.mixture.audio_db import get_augmented_noise_audios
from sonusai.mixture.audio_db import get_mixture_data
from sonusai.mixture.audio_db import get_next_noise
from sonusai.mixture.audio_db import get_raw_audios
from sonusai.mixture.audio_db import get_target_noise_audio
from sonusai.mixture.audio_db import read_audio
from sonusai.mixture.audio_db import read_raw_target_audio
from sonusai.mixture.augmentation import apply_augmentation
from sonusai.mixture.augmentation import apply_snr_gain
from sonusai.mixture.augmentation import estimate_audio_length
from sonusai.mixture.augmentation import get_augmentation_indices_for_mixup
from sonusai.mixture.augmentation import get_augmentations
from sonusai.mixture.augmentation import get_mixups
from sonusai.mixture.augmentation import get_pad_length
from sonusai.mixture.augmentation import pad_to_samples
from sonusai.mixture.balance import get_class_balancing_augmentation
from sonusai.mixture.class_count import get_class_count_from_mixids
from sonusai.mixture.class_count import get_class_count_from_mixtures
from sonusai.mixture.class_count import get_class_count_from_mrecord
from sonusai.mixture.config import get_default_config
from sonusai.mixture.config import get_max_class
from sonusai.mixture.config import get_noise_files
from sonusai.mixture.config import get_target_files
from sonusai.mixture.config import load_config
from sonusai.mixture.config import raw_load_config
from sonusai.mixture.config import tokenized_expandvars
from sonusai.mixture.config import update_config_from_file
from sonusai.mixture.config import update_config_from_hierarchy
from sonusai.mixture.config import update_truth_settings
from sonusai.mixture.constants import BIT_DEPTH
from sonusai.mixture.constants import CHANNEL_COUNT
from sonusai.mixture.constants import DEFAULT_CONFIG
from sonusai.mixture.constants import DEFAULT_FRAME_SIZE
from sonusai.mixture.constants import DEFAULT_NOISE
from sonusai.mixture.constants import FLOAT_BYTES
from sonusai.mixture.constants import RAND_PATTERN
from sonusai.mixture.constants import REQUIRED_CONFIGS
from sonusai.mixture.constants import SAMPLE_BYTES
from sonusai.mixture.constants import SAMPLE_RATE
from sonusai.mixture.constants import VALID_AUGMENTATIONS
from sonusai.mixture.constants import VALID_CONFIGS
from sonusai.mixture.constants import VALID_TRUTH_SETTINGS
from sonusai.mixture.feature import get_feature_and_truth_f
from sonusai.mixture.feature import get_feature_from_audio
from sonusai.mixture.feature import get_feature_stats
from sonusai.mixture.feature import get_ft_from_dir
from sonusai.mixture.feature import read_feature_data
from sonusai.mixture.log_duration_and_sizes import log_duration_and_sizes
from sonusai.mixture.mapped_snr_f import calculate_mapped_snr_f
from sonusai.mixture.mapped_snr_f import calculate_snr_f_statistics
from sonusai.mixture.mixdb import Augmentation
from sonusai.mixture.mixdb import Augmentations
from sonusai.mixture.mixdb import AugmentedTarget
from sonusai.mixture.mixdb import AugmentedTargets
from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MRecords
from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs
from sonusai.mixture.mixdb import TargetFile
from sonusai.mixture.mixdb import TargetFiles
from sonusai.mixture.mixdb import TruthSettings
from sonusai.mixture.mixdb import convert_json_to_mixdb
from sonusai.mixture.mixdb import convert_mixids_to_list
from sonusai.mixture.mixdb import evaluate_random_rule
from sonusai.mixture.mixdb import get_class_weights_threshold
from sonusai.mixture.mixdb import get_feature_frames_in_mrecord
from sonusai.mixture.mixdb import get_file_frame_segments
from sonusai.mixture.mixdb import get_mixid_padded_name
from sonusai.mixture.mixdb import get_mixids_data
from sonusai.mixture.mixdb import get_mixture_metadata
from sonusai.mixture.mixdb import get_mrecords_from_mixids
from sonusai.mixture.mixdb import get_samples_in_mixid
from sonusai.mixture.mixdb import get_transform_frames_in_mrecord
from sonusai.mixture.mixdb import load_mixdb
from sonusai.mixture.mixdb import new_mixdb_from_mixids
from sonusai.mixture.process import process_mixture
from sonusai.mixture.process import process_target
from sonusai.mixture.segment import Segment
from sonusai.mixture.segment import get_slices
from sonusai.mixture.segsnr import generate_segsnr
from sonusai.mixture.target_class_balancing import balance_targets
from sonusai.mixture.targets import get_augmented_target_indices_by_class
from sonusai.mixture.targets import get_augmented_target_indices_for_mixup
from sonusai.mixture.targets import get_augmented_targets
from sonusai.mixture.targets import get_target_indices_for_truth_index
from sonusai.mixture.targets import get_truth_indices_for_target
from sonusai.mixture.truth import generate_truth
from sonusai.mixture.truth import get_truth_indices_for_mixid
from sonusai.mixture.truth import truth_function
from sonusai.mixture.truth import truth_reduction



from dataclasses import dataclass
from typing import List

from onnxruntime import InferenceSession


def replace_stateful_grus(keras_model, onnx_model):
    """Replace stateful GRUs with custom layers."""
    from tensorflow.keras.layers import GRU

    stateful_gru_names = []
    for i in range(len(keras_model.layers)):
        layer = keras_model.layers[i]
        if isinstance(layer, GRU):
            if layer.stateful:
                stateful_gru_names.append(layer.name)

    for node_index in range(len(onnx_model.graph.node)):
        node = onnx_model.graph.node[node_index]
        replace = False
        if node.op_type == 'GRU':
            for i in node.input:
                for n in stateful_gru_names:
                    if n in i:
                        replace = True
        if node.name in stateful_gru_names or replace:
            node.op_type = 'SGRU'

    return onnx_model


def add_sonusai_metadata(model,
                         is_flattened: bool = True,
                         has_timestep: bool = True,
                         has_channel: bool = False,
                         is_mutex: bool = True,
                         feature: str = ''):
    """Add SonusAI metadata to ONNX model.
      model           keras model
      is_flattened    model feature data is flattened
      has_timestep    model has timestep dimension
      has_channel     model has channel dimension
      is_mutex        model label output is mutually exclusive
      feature         model feature type
    """
    is_flattened_flag = model.metadata_props.add()
    is_flattened_flag.key = 'is_flattened'
    is_flattened_flag.value = str(is_flattened)

    has_timestep_flag = model.metadata_props.add()
    has_timestep_flag.key = 'has_timestep'
    has_timestep_flag.value = str(has_timestep)

    has_channel_flag = model.metadata_props.add()
    has_channel_flag.key = 'has_channel'
    has_channel_flag.value = str(has_channel)

    is_mutex_flag = model.metadata_props.add()
    is_mutex_flag.key = 'is_mutex'
    is_mutex_flag.value = str(is_mutex)

    feature_flag = model.metadata_props.add()
    feature_flag.key = 'feature'
    feature_flag.value = str(feature)

    return model


@dataclass(frozen=True)
class SonusAIMetaData:
    input_shape: List[int]
    output_shape: List[int]
    flattened: bool
    timestep: bool
    channel: bool
    mutex: bool
    feature: str


def get_sonusai_metadata(model: InferenceSession) -> SonusAIMetaData:
    m = model.get_modelmeta().custom_metadata_map
    return SonusAIMetaData(input_shape=model.get_inputs()[0].shape,
                           output_shape=model.get_outputs()[0].shape,
                           flattened=m['is_flattened'] == 'True',
                           timestep=m['has_timestep'] == 'True',
                           channel=m['has_channel'] == 'True',
                           mutex=m['is_mutex'] == 'True',
                           feature=m['feature'])



from typing import List

import numpy as np

from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MixtureDatabase


def _process_target_audio(mixdb: MixtureDatabase,
                          mrecord: MRecord,
                          raw_target_audios: List[np.ndarray]) -> List[np.ndarray]:
    """Apply augmentation and update target metadata."""
    from sonusai import SonusAIError
    from sonusai.mixture import pad_to_samples
    from sonusai.mixture.augmentation import apply_augmentation

    target_file_index = mrecord.target_file_index
    target_augmentation_index = mrecord.target_augmentation_index
    if len(target_file_index) != len(target_augmentation_index):
        raise SonusAIError('target_file_index and target_augmentation_index are not the same length')

    targets = []
    mrecord.target_gain = []
    for idx in range(len(target_file_index)):
        target_augmentation = mixdb.target_augmentations[target_augmentation_index[idx]]

        targets.append(apply_augmentation(audio_in=raw_target_audios[target_file_index[idx]],
                                          augmentation=target_augmentation,
                                          length_common_denominator=mixdb.feature_step_samples))

        # target_gain is used to back out the gain augmentation in order to return the target audio
        # to its normalized level when calculating truth.
        if target_augmentation.gain is not None:
            target_gain = 10 ** (target_augmentation.gain / 20)
        else:
            target_gain = 1
        mrecord.target_gain.append(target_gain)

    mrecord.samples = max([len(item) for item in targets])

    for idx in range(len(targets)):
        targets[idx] = pad_to_samples(audio_in=targets[idx], samples=mrecord.samples)

    return targets


def process_target(mixdb: MixtureDatabase,
                   mrecord: MRecord,
                   raw_target_audios: List[np.ndarray]) -> MRecord:
    """Apply augmentation and update target metadata."""
    _process_target_audio(mixdb, mrecord, raw_target_audios)
    return mrecord


def process_mixture(mixdb: MixtureDatabase,
                    mrecord: MRecord,
                    raw_target_audios: List[np.ndarray],
                    augmented_noise_audios: List[List[np.ndarray]]) -> (MRecord,
                                                                        np.ndarray,
                                                                        np.ndarray,
                                                                        np.ndarray,
                                                                        np.ndarray):
    from sonusai.mixture import apply_snr_gain
    from sonusai.mixture import get_next_noise
    from sonusai.mixture import get_class_count_from_mrecord

    targets = _process_target_audio(mixdb=mixdb,
                                    mrecord=mrecord,
                                    raw_target_audios=raw_target_audios)

    noise = augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
    noise, _ = get_next_noise(offset_in=mrecord.noise_offset,
                              length=mrecord.samples,
                              audio_in=noise)

    mrecord = _update_mixture_gains(mixdb=mixdb,
                                    mrecord=mrecord,
                                    target_audios=targets,
                                    noise_audio=noise)

    targets, noise = apply_snr_gain(mrecord=mrecord, targets=targets, noise=noise)
    target = sum(targets)
    mixture = target + noise

    mrecord.class_count, truth_t = get_class_count_from_mrecord(mixdb=mixdb,
                                                                mrecord=mrecord,
                                                                target_audios=targets,
                                                                noise_audio=noise)

    return mrecord, mixture, target, noise, truth_t


def _update_mixture_gains(mixdb: MixtureDatabase,
                          mrecord: MRecord,
                          target_audios: List[np.ndarray],
                          noise_audio: np.ndarray) -> MRecord:
    from sonusai.utils import dbfs_to_sample

    target_audio = np.single(sum(target_audios))

    if mrecord.snr < -96:
        # Special case for zeroing out target data
        mrecord.target_snr_gain = 0
        mrecord.noise_snr_gain = 1
        mrecord.class_count = [0] * mixdb.num_classes
        # Setting target_gain to zero will cause the truth to be all zeros.
        mrecord.target_gain = [0] * len(mrecord.target_gain)
    elif mrecord.snr > 96:
        # Special case for zeroing out noise data
        mrecord.target_snr_gain = 1
        mrecord.noise_snr_gain = 0
    else:
        target_energy = np.mean(np.square(target_audio))
        noise_energy = np.mean(np.square(np.single(noise_audio)))
        if noise_energy == 0:
            noise_gain = 1
        else:
            noise_gain = np.sqrt(target_energy / noise_energy) / 10 ** (mrecord.snr / 20)

        # Check for noise_gain > 1 to avoid clipping
        if noise_gain > 1:
            mrecord.target_snr_gain = 1 / noise_gain
            mrecord.noise_snr_gain = 1
        else:
            mrecord.target_snr_gain = 1
            mrecord.noise_snr_gain = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = target_audio * mrecord.target_snr_gain
    gain_adjusted_noise_audio = np.single(noise_audio) * mrecord.noise_snr_gain
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio
    if any(abs(mixture_audio) >= 32768):
        # Clipping occurred; lower gains to bring audio within int16 bounds
        gain_adjustment = dbfs_to_sample(-0.25) / max(abs(mixture_audio))
        mrecord.target_snr_gain *= gain_adjustment
        mrecord.noise_snr_gain *= gain_adjustment

    return mrecord



"""sonusai predict

usage: predict [-hvr] (-m MODEL) (-d INPUT) [-o OUTPUT]

options:
    -h, --help
    -v, --verbose               Be verbose.
    -m MODEL, --model MODEL     Trained ONNX model file.
    -d INPUT, --input INPUT     Input data.
    -o OUTPUT, --output OUTPUT  Output directory.
    -r, --reset                 Reset model between each file.

Run prediction on a trained ONNX model using SonusAI genft data and/or WAV data.

Inputs:
    MODEL       A SonusAI trained ONNX model file.
    INPUT       The input data can be a file, directory, or glob of genft H5 files and/or WAV files.

Outputs:
    OUTPUT/     A directory containing:
                    <id>.h5
                        dataset:    predict
                    predict.log

Results are written into subdirectory <MODEL>-<TIMESTAMP> unless OUTPUT is specified.

"""
from sonusai import logger


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    input_name = args['--input']
    output_dir = args['--output']
    reset = args['--reset']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import h5py
    import onnxruntime as rt
    import numpy as np

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import get_frames_per_batch
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import read_audio
    from sonusai.utils import braced_glob
    from sonusai.utils import create_ts_name
    from sonusai.utils import get_sonusai_metadata
    from sonusai.utils import reshape_inputs
    from sonusai.utils import reshape_outputs

    if output_dir is None:
        output_dir = create_ts_name('opredict')

    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'onnx_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('onnx_predict')

    # model = Predict(model_name)
    model = rt.InferenceSession(model_name, providers=['CPUExecutionProvider'])
    output_names = [n.name for n in model.get_outputs()]
    input_names = [n.name for n in model.get_inputs()]
    model_metadata = get_sonusai_metadata(model)

    batch_size = model_metadata.input_shape[0]
    if model_metadata.timestep:
        timesteps = model_metadata.input_shape[1]
    else:
        timesteps = 0
    num_classes = model_metadata.output_shape[-1]

    frames_per_batch = get_frames_per_batch(batch_size, timesteps)

    logger.info('')
    logger.info(f'feature       {model_metadata.feature}')
    logger.info(f'num_classes   {num_classes}')
    logger.info(f'batch_size    {batch_size}')
    logger.info(f'timesteps     {timesteps}')
    logger.info(f'flatten       {model_metadata.flattened}')
    logger.info(f'add1ch        {model_metadata.channel}')
    logger.info(f'truth_mutex   {model_metadata.mutex}')
    logger.info(f'input_shape   {model_metadata.input_shape}')
    logger.info(f'output_shape  {model_metadata.output_shape}')
    logger.info('')

    if isdir(input_name):
        input_name = join(input_name, '*.{h5,wav}')

    files = [x for x in braced_glob(input_name) if splitext(x)[1] == '.h5']

    # Convert WAV to feature data
    wav_files = [x for x in braced_glob(input_name) if splitext(x)[1] == '.wav']
    for file in wav_files:
        audio = read_audio(file)
        data = get_feature_from_audio(audio=audio, feature=model_metadata.feature)

        output_base = basename(splitext(file)[0])
        output_name = join(output_dir, output_base)
        if output_base + '.h5' in [basename(x) for x in files]:
            # If an H5 file of the same name as the WAV file already exists, then append '_wav'
            # to the output file name.
            output_name += '_wav'
        output_name += '.h5'
        files.append(output_name)
        # Save the newly created feature
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=data, dtype=np.single)

    files = sorted(files)
    logger.info(f'Found {len(files)} files from {input_name}')

    if reset:
        # reset mode cycles through each file one at a time
        for file in files:
            # Read in H5
            with h5py.File(file, 'r') as f:
                data = np.array(f['feature'])

            # Pad with zeros in order to create an entire batch of data
            frames = data.shape[0]
            padding = frames_per_batch - data.shape[0] % frames_per_batch
            data = np.pad(array=data,
                          pad_width=((0, padding), (0, 0), (0, 0)),
                          mode='constant',
                          constant_values=0)
            data, _, _, _, _, _ = reshape_inputs(feature=data,
                                                 batch_size=batch_size,
                                                 timesteps=timesteps,
                                                 flatten=model_metadata.flattened,
                                                 add1ch=model_metadata.channel)
            sequences = data.shape[0] // model_metadata.input_shape[0]
            data = np.reshape(data, [sequences, *model_metadata.input_shape])

            # Create new model for each file; simulates reset
            # model = Predict(model_name)
            # predict = model.execute(data)
            predict = []
            for s in range(sequences):
                model = rt.InferenceSession(model_name, providers=['CPUExecutionProvider'])
                predict.append(model.run(output_names, {input_names[0]: data[s]}))
            predict = np.vstack(predict)
            predict, _, _ = reshape_outputs(predict=predict, timesteps=timesteps)
            predict = predict[:frames, :]

            output_name = join(output_dir, basename(file))
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset(name='predict', data=predict, dtype=np.single)
    else:
        # Read in H5s
        data = []
        file_indices = []
        total_frames = 0
        for file in files:
            with h5py.File(file, 'r') as f:
                current_data = np.array(f['feature'])
                current_frames = current_data.shape[0]
                data.append(current_data)
                file_indices.append(slice(total_frames, total_frames + current_frames))
                total_frames += current_frames
        data = np.vstack([data[i] for i in range(len(data))])

        # Pad with zeros in order to create an entire batch of data
        frames = data.shape[0]
        padding = frames_per_batch - frames % frames_per_batch
        data = np.pad(array=data,
                      pad_width=((0, padding), (0, 0), (0, 0)),
                      mode='constant',
                      constant_values=0)

        data, _, _, _, _, _ = reshape_inputs(feature=data,
                                             batch_size=batch_size,
                                             timesteps=timesteps,
                                             flatten=model_metadata.flattened,
                                             add1ch=model_metadata.channel)
        sequences = data.shape[0] // model_metadata.input_shape[0]
        data = np.reshape(data, [sequences, *model_metadata.input_shape])

        # predict = model.execute(data)
        predict = []
        for s in range(sequences):
            predict.append(model.run(output_names, {input_names[0]: data[s]}))
        predict = np.vstack(predict)
        predict, _, _ = reshape_outputs(predict=predict, timesteps=timesteps)
        predict = predict[:frames, :]

        # Write data to separate files
        for idx, file in enumerate(files):
            output_name = join(output_dir, basename(file))
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset('predict', data=predict[file_indices[idx]], dtype=np.single)

    logger.info(f'Saved results to {output_dir}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from dataclasses import dataclass

import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureIDs


class KerasFromMixDB(Sequence):
    """Generates data for Keras from a mixture database"""

    @dataclass(frozen=True)
    class BatchParams:
        mixids: MixtureIDs
        offset: int
        extra: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: MixtureIDs,
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        from sonusai.mixture import convert_mixids_to_list
        from sonusai.mixture import get_augmented_noise_audios
        from sonusai.mixture import get_feature_stats
        from sonusai.mixture import get_raw_audios

        self.mixdb = mixdb
        self.mixids = convert_mixids_to_list(self.mixdb, mixids)
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle
        self.raw_target_audios = get_raw_audios(self.mixdb)
        self.augmented_noise_audios = get_augmented_noise_audios(self.mixdb)

        fs = get_feature_stats(feature=self.mixdb.feature,
                               num_classes=self.mixdb.num_classes,
                               frame_size=self.mixdb.frame_size)

        self.stride = fs.stride
        self.num_bands = fs.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mixtures = None
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""
        from sonusai.utils import reshape_inputs
        from sonusai.utils import p_map

        batch_params = self.batch_params[batch_index]

        result = p_map(self._fetch_mixtures, batch_params.mixids)

        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth, _, _, _, _ = reshape_inputs(feature=feature,
                                                    truth=truth,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        import random

        if self.shuffle:
            random.shuffle(self.mixids)
            self.initialize_mixtures()

    def _fetch_mixtures(self, mixid: int) -> (np.ndarray, np.ndarray, np.ndarray):
        from sonusai.genft import genft

        return genft(mixdb=self.mixdb,
                     mixids=mixid,
                     raw_target_audios=self.raw_target_audios,
                     augmented_noise_audios=self.augmented_noise_audios,
                     compute_segsnr=False)

    def _initialize_mixtures(self) -> None:
        from sonusai.data_generator.utils import get_frames_per_batch_from_mixdb
        from sonusai.mixture import get_mrecords_from_mixids

        self.mixtures = get_mrecords_from_mixids(mixdb=self.mixdb, mixids=self.mixids)
        (frames_per_batch,
         self.total_batches) = get_frames_per_batch_from_mixdb(mixdb=self.mixdb,
                                                               mrecords=self.mixtures,
                                                               batch_size=self.batch_size,
                                                               timesteps=self.timesteps)

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        for m in range(len(self.mixids)):
            current_frames = self.mixtures[m].samples // self.mixdb.feature_step_samples
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                self.batch_params.append(self.BatchParams(mixids=self.mixids[start_mixture_index:m + 1],
                                                          offset=offset,
                                                          extra=extra))
                if extra == 0:
                    start_mixture_index = m + 1
                    offset = 0
                else:
                    start_mixture_index = m
                    offset = current_frames - extra
                cumulative_frames = extra



"""sonusai

usage: sonusai [--version] [--help] <command> [<args>...]

The sonusai commands are:
   evaluate         Evaluate model performance
   genft            Generate feature and truth data
   genmix           Generate mixture and truth data
   genmixdb         Generate a mixture database
   gentcst          Generate target configuration from a subdirectory tree
   keras_predict    Run Keras predict on a trained model
   keras_train      Train a model using Keras
   keras_onnx       Convert a trained Keras model to ONNX
   lsdb             List information about a mixture database
   mkwav            Make WAV files from a mixture database
   onnx_predict     Run ONNX predict on a trained model
   plot             Plot mixture data
   tplot            Plot truth data
   vars             List custom SonusAI variables

Aaware Sound and Voice Machine Learning Framework. See 'sonusai help <command>'
for more information on a specific command.

"""

from sonusai import logger


def main():
    from subprocess import call

    from docopt import docopt

    import sonusai
    from sonusai import logger
    from sonusai.utils import trim_docstring

    commands = (
        'evaluate',
        'genft',
        'genmix',
        'genmixdb',
        'gentcst',
        'keras_predict',
        'keras_train',
        'keras_onnx',
        'lsdb',
        'mkwav',
        'onnx_predict',
        'plot',
        'tplot',
        'vars',
    )

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    command = args['<command>']
    argv = args['<args>']

    if command == 'help':
        if not argv:
            exit(call(['sonusai', '-h']))
        elif argv[0] in commands:
            exit(call(['python', f'{sonusai.BASEDIR}/{argv[0]}.py', '-h']))
        else:
            logger.error(f"{argv[0]} is not a SonusAI command. See 'sonusai help'.")
            raise SystemExit(1)
    elif command in commands:
        exit(call(['python', f'{sonusai.BASEDIR}/{command}.py'] + argv))

    logger.error(f"{command} is not a SonusAI command. See 'sonusai help'.")
    raise SystemExit(1)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)
